require 'weaver/core/logging'
require 'weaver/dsl/weaver_parser'
require 'json'
require 'weaver/utils/file_helper'

module Weaver
  # Weaver files may include +chef_automation+ elements to trigger deployment using {Chef}[http://www.opscode.com/chef/].
  #
  # Weaver includes a tool to create chef-solo configuration from a Weaver file, 
  # a tool to introspect Chef cookbooks to create Weaver descriptions, and a tool to validate Weaver descriptions of Chef against drift.
  #
  # @see ChefAutomationHandler
  # @see ChefNodeInspector
  # @see RoleComparator
  module Chef

    # Enhanced, recursive mini-DSL processor which is able to parse multi-level 
    # nested Chef attributes, e.g., <code>default['apache']['worker']['startservers'] = 4</code>
    # 
    # Author: Waldemar Hummer
    # Author: Ed Snible
    class DefaultProcessorRecursive < Hash
      def initialize(node_values = {})
        @user_values = Hash.new.merge(node_values) if node_values
      end
      
      # Parse Chef's +default+ statement 
      def default(arg = nil)
        if arg
          if !self.include?(arg)
            store(arg, self.class.new)
          end
          return self[arg]
        end
        return self
      end
      
      # Parse Chef's +set+ statement 
      def set(arg = nil)
        if arg
          store(arg, self.class.new)
          return self[arg]
        end
        return self
      end
      
      # Parse Chef's +node+ statement 
      def node(arg = nil)
        new_hash = self.class.new
        new_hash.merge!(@user_values["node"]) if @user_values["node"]
        return new_hash
      end
      
      # Parse Chef's +kernel+ statement 
      def kernel(arg = nil)
        new_hash = self.class.new
        new_hash.merge!(@user_values["node"]) if @user_values["node"]
        return new_hash
      end
      
      # @param arg [String] Chef parameter name
      # @return +true+ if the key has a known value
      def attribute?(arg)
        return !@user_values[arg].nil?
      end
      
      def method_missing(sym, *arguments, &block)
        if key?(sym)
          return fetch(sym)
        end
        if key?(sym.to_s)
          return fetch(sym.to_s)
        end
        
        if ["platform", "platform_version", "ipaddress", "fqdn", 
          "macaddress", "hostname", "domain", "recipes", "roles", 
          "ohai_time", 'python'].include?(sym.to_s)
          return @user_values[sym.to_s]
        else
          Weaver::Core::Logging::get_logger.warn "Missing method #{sym} when trying to parse Chef node attributes"
#          raise "Missing method #{sym} when trying to parse Chef node attributes"
        end
      end
    end


    #
    # Introspects \Chef role files and generates a \Weaver representation that
    # contains a +chef_automation+ type. This file can then be included 
    # into a +node+ for deploying certain aspects of an application.   
    #
    class ChefNodeInspector
      include Weaver::Core

      @@DEFAULT_PROCESSOR_CLASS = Weaver::Chef::DefaultProcessorRecursive
      
#      # When scraping Chef configuration, pretend to be a machine that looks like this
#      @@DEFAULT_PSUEDO_OHAI_ATTRIBUTES = {
#        "node" => {
#          "platform" => "fedora", 
#          "kernel" => { "machine" => "i386" },
#          "platform_version" => "14",
#          "cpu" => { "total" => "1" }
#        }, 
#        "platform" => "fedora",
#        "tomcat5" => {}
#      }
       
      # An array of VM Attributes often accessed by Chef attribute .rbs, 
      # such as node['platform'] and node['kernel']['machine']
      # to be offered to detect which attributes are platform-specific 
      @@SIMULATED_VM_ATTRIBUTES = [
        { "node" => {
          "platform" => "fedora", 
          "kernel" => { "machine" => "i386" },
          "platform_version" => "14",
          "cpu" => { "total" => "1" }
        }, 
        "platform" => "fedora",
        "tomcat5" => {} },
        { "node" => {
          "platform" => "ubuntu", 
          "kernel" => { "machine" => "x86_64" },
          "platform_version" => "14",
          "cpu" => { "total" => "1" }
        }, 
        "platform" => "ubuntu",
        "tomcat5" => {} },
        { "node" => {
          "platform" => "amazon", 
          "kernel" => { "machine" => "x86_64" },
          "platform_version" => "1",
          "cpu" => { "total" => "2" }
        }, 
        "platform" => "amazon",
        "tomcat5" => {} },
        ] 

      @processor = nil
      
      # @private
      # Optional Weaver ezcrypto key file filename to override ENV['KEY_FILE'].  Deprecated.
      attr_writer :keyfile

      private
      
      # @private
      # :stopdoc:
      # Format of a Recipe name      
      # \x2D allows - characters
      RECIPE_REGEXP = /([a-z_A-Z0-9\x2D]+)(::([a-z_A-Z0-9\x2D]+))?/
      # :startdoc:
      
      public
      
      # @param chef_paths [String[]] \Chef directories on local filesystem
      # @param error_reporting - a <code>Proc { |level, message| }</code> for reporting progress.
      def initialize(chef_paths, &error_reporting)
        @chef_paths = chef_paths
#        @processor = processor
        raise "Invalid 'error_reporting'" if error_reporting.nil?
        @error_reporting = error_reporting
        @keyfile = nil
        @scraped_recipes = Set.new
      end
      
      # @private
      # :stopdoc:
      # Given a JSON description of a node, return a list of qualified recipe names
      def get_node_recipes(json_node_file)
        full_filename = get_filename("nodes/" + json_node_file)
        raise WeaverError.new("File not found: 'nodes/#{json_node_file}' in '#{@chef_paths.inspect}'") unless full_filename
        
        node = JSON.parse(File.new(full_filename).read())
          
        recipies = node['run_list'].map { |item| item[/recipe\[(.*)\]/, 1] }.select { |item| item }
          
        roles = node['run_list'].map { |item| item[/role\[(.*)\]/, 1] }.select { |item| item }
        roles.each do |role|
          recipies.concat(get_recipes(get_filename("roles/" + role + ".rb")))
        end
          
        return recipies
      end
      
      # @private
      # Given a \Chef description of a role, return a list of qualified recipe names
      def get_role_recipes(ruby_role_file)
        full_filename = get_filename("roles/" + ruby_role_file)
        raise WeaverError.new("File not found: 'roles/#{ruby_role_file}' in '#{@chef_paths.inspect}'") unless full_filename
        return get_recipes(full_filename)
      end
      
      # @private
      # Given a \Chef cookbook name, scrape the default parameter files and return a Hash of Hashes, recipe->param->default
      def get_defaults(cookbook_name, recipe_name)
        @@logger.debug "Called with #{cookbook_name}::#{recipe_name}"
        if cookbook_name.nil?
          @error_reporting.call(Logger::WARN, "Unknown cookbook #{cookbook_name.inspect}")
          return {}
        end
        
        default_file = get_filename("cookbooks/" + cookbook_name + "/attributes/#{recipe_name}.rb")
        
        # If there is no file with the expected name use the default.rb
        if default_file.nil?
          default_file = get_filename("cookbooks/" + cookbook_name + "/attributes/default.rb")
          @@logger.debug "Couldn't find cookbooks/#{cookbook_name}/attributes/#{recipe_name}.rb, using default.rb" unless default_file.nil? 
        end
        
        # If there is no file with the expected name or default name use the cookbook name
        if default_file.nil?
          default_file = get_filename("cookbooks/#{cookbook_name}/attributes/#{cookbook_name}.rb")
          @@logger.debug "Couldn't find cookbooks/#{cookbook_name}/attributes/#{recipe_name}.rb, using #{cookbook_name}.rb" unless default_file.nil? 
        end
        
        if default_file.nil?
          @error_reporting.call(Logger::INFO, "Unknown default file #{("cookbooks/" + cookbook_name + "/attributes/#{recipe_name}.rb").inspect}")
          return {}
        end
        
        default_expr = File.new(default_file).read()
        return multiparse_defaults(default_expr, default_file)
      end
      
      # @private
      # Given a recipe, get a list of parameter names by scanning the contents looking for node[].
      def scrape_direct_param_usage(ruby_recipe)
        contents = File.new(ruby_recipe).read()
        
        usage = []
          
        # Use a regular expression to search for node['<STRING>']['<STRING'> and node["<STRING>"]["<STRING>"]
        # Note that there can be more 2 occurrences of ['<STRING>'] or ["<STRING">]
        # Also search for node[:<SYMBOL>][:<SYMBOL].  Also note an example with "ARCH = node[:java]['arch']"
        contents.scan(/(node(\[[:'"][^\]]*\]){2,})/) do |m|
          usage.push m[0].sub(/node\[[:'"]/, "").gsub(/['"]?\]\[[:'"]/,".").sub(/['"]?\]$/,"");
        end
        
        # Process +include_recipe+ Chef directive
        contents.scan(/include_recipe *['"](.*)['"]/) do |md|
          included_recipe_name = md[0]
          
          if @scraped_recipes.member?(included_recipe_name)
            @error_reporting.call(Logger::WARN, "Possibly recursive recipe reference #{included_recipe_name.inspect}")
            return usage.uniq
          end
          @scraped_recipes.add(included_recipe_name)
          
#          @error_reporting.call(Logger::WARN, "Recipe file #{ruby_recipe.inspect} uses include_recipe #{included_recipe_name}")
          included_filename = get_filename(get_relative_recipe_files([included_recipe_name])[0])
          if included_filename.nil?
            @error_reporting.call(Logger::WARN, "Recipe file #{ruby_recipe.inspect} includes unknown recipe #{included_recipe_name}")
          else
#            @error_reporting.call(Logger::INFO, "Recipe #{ruby_recipe.inspect} includes #{included_filename}")
            usage.concat(scrape_param_usage(included_filename))
          end
        end
          
        return usage.uniq
      end
      
      # @private
      # Given a recipe, get a list of source files.
      def scrape_source_usage(ruby_recipe)
        contents = File.new(ruby_recipe).read()
        #   source 'mysql/create_db.sql.erb'

        usage = []
        contents.scan(/source ['"](.*)['"]/) { |fname, *dummy| usage.push fname }
        return usage.uniq
      end
          
      # @private
      # Given a recipe, get a list of parameter names and defaults by scanning the contents.
      def scrape_param_usage(ruby_recipe)
        @@logger.debug "Inspecting parameter usage by #{ruby_recipe}"
        
        unless File.exists?(ruby_recipe)
          @error_reporting.call(Logger::WARN, "Recipe file not found: #{ruby_recipe.inspect}")
          return []
        end
          
        retval = scrape_direct_param_usage(ruby_recipe)
        
        # sources mentioned in the chef recipe may either be stored in the files directory
        # (included with cookbook_file) or in the template directory (included with
        # template), currently we look in either directory, however, instead of using a regex
        # we should parse the recipe file in the future
        files_directory = File.expand_path("../files/default", File.dirname(ruby_recipe))
        template_directory = File.expand_path("../templates/default", File.dirname(ruby_recipe))
        scrape_source_usage(ruby_recipe).each do |filename|
          files_filename = File.expand_path(filename, files_directory)        
          template_filename = File.expand_path(filename, template_directory)        

          if File.exists?(template_filename) && File.ftype(template_filename)=='file' && File.readable?(template_filename)
            retval.concat(scrape_direct_param_usage(template_filename))
          elsif filename =~ /#\{/ 
            @error_reporting.call(Logger::WARN, "Skipping included calculated filename #{filename.inspect} used by #{ruby_recipe.inspect}")
          elsif !File.exists?(files_filename)
            @error_reporting.call(Logger::WARN, "Recipe file not found: neither #{files_filename.inspect} nor #{template_filename.inspect}")
          end          
        end
        
        return retval.uniq
      end
      
#      # Given an array of fully-qualified recipes, get a list of cookbooks needed by those recipes
#      def get_cookbooks(recipes)
#        return recipes.map { |qualified_recipe| qualified_recipe[RECIPE_REGEXP, 1] }.uniq
#      end
      
      # @private
      # Given an array of fully-qualified recipes, get a list of Ruby .rb files that implement those recipes
      def get_relative_recipe_files(recipes)
        return recipes.map { |qualified_recipe| [qualified_recipe[RECIPE_REGEXP, 1], qualified_recipe[RECIPE_REGEXP, 3] || 'default'] }.
        map { |cookbook, recipe| "cookbooks/" + cookbook.to_s + '/recipes/' + recipe.to_s + '.rb' }
      end
      # :startdoc:

      # Given a \Chef role .rb file, return a {Weaver::Core::Root} model resembling the below.
      # * +ruby_role_file+ - a \Chef .rb file
      # * +destination_dir+ - the destination for serializing the returned root (used to create a relative reference to the role file)
      #     # app_server_role.weaver
      #     # Generated by weaver chef import from app_server_role.rb
      #     chef_automation (:app_server_role) {
      #       role "jke_application"<br>
      #       property :jke_db_hostname => "localhost", :validators => [value_type(:string)], :name => "jke.db_hostname"
      #       property :jke_db_port => 50001, :validators => [value_type(:numeric)], :name => "jke.db_port"
      #       property :jke_db_name => "JKEDB", :validators => [value_type(:string)],  :name => "jke.db_name"
      #       # etc
      #     }
      def model_role(ruby_role_file, destination_dir)
        basename = File.basename(ruby_role_file, ".rb")
        recipes = get_role_recipes(ruby_role_file)
        # e.g. [ 'mysql::server', 'jke::create_database_mysql' ]
        @@logger.debug "recipes: #{recipes.inspect}"
        
#        cookbooks = get_cookbooks(recipes)
#        @@logger.debug "cookbooks: #{cookbooks.inspect}"
        all_defaults = get_all_defaults(recipes)
        @@logger.debug "all defaults: #{all_defaults.inspect}"
        process_overrides(ruby_role_file, all_defaults)
        ruby_recipes = get_relative_recipe_files(recipes)
        @@logger.debug "ruby recipes: #{ruby_recipes.inspect}"
        param_defaults = get_param_defaults(ruby_recipes, all_defaults)
        @@logger.debug "parameter defaults: #{param_defaults.inspect}"
        
        root = Weaver::Dsl::WeaverParser.new().parse(nil)
        automation = root.prototype_instance(:chef_automation, basename.symbolize, root, nil)
        automation.prefix_comments.push "#{basename}.weaver"
        automation.prefix_comments.push "Generated by weaver chef import from #{ruby_role_file}"
        root.automation[automation.id] = automation
        
        automation.role = basename
        full_filename = get_filename("roles/" + ruby_role_file)
#        automation.rolefile = Weaver::Utils::FileHelper.relativize(full_filename, File.expand_path(destination_dir))
        defaults_to_params(param_defaults, automation)
        
        return root
      end
      
      # Given a \Chef role .rb file, apply the overrides to the defaults
      # * +ruby_role_file+ - a \Chef .rb file
      # * +all_defaults+ - Hash of default values
      def process_overrides(ruby_role_file, all_defaults)
        # Re-read the role file, this time looking for overrides
        full_filename = get_filename("roles/" + ruby_role_file)
        raise WeaverError.new("File not found: 'roles/#{ruby_role_file}' in '#{@chef_paths.inspect}'") unless full_filename
        role_processor = RoleProcessor.new()
        role_processor.instance_eval(File.new(full_filename).read(), ruby_role_file)

        defaults = role_processor.default_attributes()
        recursive_merge!(all_defaults, defaults) unless defaults.nil?
        overrides = role_processor.override_attributes()
        recursive_merge!(all_defaults, overrides) unless overrides.nil?
      end
      
      # Given a \Chef node .json file, return a Weaver::Core::Root with attributes 
      # representing the recipes on the node.
      def model_node(json_node_file)
        recipes = get_node_recipes(json_node_file)
        # [ 'mysql::server', 'jke::create_database_mysql' ]
        
#        cookbooks = get_cookbooks(recipes)
        all_defaults = get_all_defaults(recipes)
        ruby_recipes = get_relative_recipe_files(recipes)
        param_defaults = get_param_defaults(ruby_recipes, all_defaults)            
        
        root = Root.new(:dummy, nil)
        defaults_to_params(param_defaults, root)
        
        chef_node = Attribute.new(:chef_node, root)
        chef_node.name = "CHEF_NODE"
        chef_node.value = json_node_file[/([a-z0-9_]+)_node\.json/, 1]
        chef_node.prefix_comments.push "Represents Chef node " + json_node_file
        root.amd[chef_node.id] = chef_node
        
        deps_file_url = Attribute.new(:deps_file_url, root)
        deps_file_url.mutable = true
        deps_file_url.name = "DEPS_FILE_URL"
        deps_file_url.prefix_comments.push ":deps_file_url must be set by the build to point to a text file telling which cookbooks to bootstrap"
        deps_file_url.add_validator(:not_empty)
        root.amd[deps_file_url.id] = deps_file_url

        used_parameters = filter_parameters(all_defaults, param_defaults.keys)
        weaverized_parameters = weaverize(used_parameters)
        chef_node_attr = Attribute.new(:chef_node_attr, root)
        chef_node_attr.name = "CHEF_NODE_ATTR"
        chef_node_attr.value = UnquotedString.new("late_binding { #{weaverized_parameters.inspect} }")
        chef_node_attr.prefix_comments.push "Mechanism by which devOps Chef-solo passes values to the node" 
        root.amd[chef_node_attr.id] = chef_node_attr
        
        return root
      end
            
      private
      
      def recursive_merge!(hash1, hash2)
        hash1.merge!(hash2) { |_key, oldval, newval| (oldval.kind_of?(Hash) && newval.kind_of?(Hash)) ? recursive_merge!(oldval, newval) : newval }
      end

      # Supress the Ruby error 'NameError: uninitialized constant'
      # Note that we do this in a non-threadsafe manner; it is best not to expect other code
      # to use const_missing during parsing of Chef methods.
      # This is a work-around so that we don't fail to parse attribute files containing code
      # such as
      #    ::Chef::Node.send(:include, Opscode::OpenSSL::Password)
      # This code doesn't help us process that code, only keeps us from failing the ::Chef lookup.
      # See #enable_const_missing
      def disable_const_missing
        Object.class_exec {
          class << self

            def const_missing_always_succeed(sym)
              return Module.new() do
                def self.const_missing(sym)
                  return self
                end
              end
            end
            
            alias_method :const_missing_orig, :const_missing
            alias_method :const_missing, :const_missing_always_succeed
          end
        }
      end

      # See #disable_const_missing
      def enable_const_missing
        Object.class_exec {
          class << self
            alias_method :const_missing, :const_missing_orig
            undef_method :const_missing_orig
          end
        }
      end
      
      # Return +true+ if +attr+'s value should be encrypted
      # See Weaver::Core::EncryptedValue
      def should_encrypt?(attr, val)
        # return false unless val.kind_of?(String)
        return /password/.match(attr.id.to_s) || /passwd/.match(attr.id.to_s) if val.nil? || val.kind_of?(String)
        return false
      end
      
      def get_filename(filename)
        @chef_paths.each do |path|
          candidate = File.expand_path(filename, path)
          return candidate if File.exist?(candidate)
        end
        
        # No such file
        @@logger.info "File not found '#{filename}'"
        return nil
      end

      def get_recipes(ruby_role_file)
        unless File.exists?(ruby_role_file)
          @error_reporting.call(Logger::WARN, "Role file not found: #{ruby_recipe.inspect}")
          return []
        end
          
        role_processor = RoleProcessor.new()
        role_processor.instance_eval(File.new(ruby_role_file).read(), ruby_role_file)
        retval = []
        role_processor.run_list().each do |item|
          recipe_md = /recipe\[(.*)\]/.match(item)
          retval << recipe_md[1] unless recipe_md.nil?
          role_md = /role\[(.*)\]/.match(item)
          retval.concat(get_role_recipes("#{role_md[1]}.rb")) unless role_md.nil?
        end
       
        return retval
      end
      
      # Given an array of recipes names, e.g. [ 'jke', 'java::jdk' ]
      # generate a Hash containing the default values of the parameters
      def get_all_defaults(recipes)
        
        disable_const_missing()
                  
        all_defaults = Hash.new { |h, k| h[k] = {} }
        recipes.map { |qualified_recipe| [qualified_recipe[RECIPE_REGEXP, 1], qualified_recipe[RECIPE_REGEXP, 3] || 'default'] }.each do |cookbook_name, recipe_name|
          @@logger.debug "Merging cookbook #{cookbook_name} recipe #{recipe_name}"
          defaults = get_defaults(cookbook_name, recipe_name)
          @@logger.debug "Defaults to merge are: #{defaults.inspect}"
          all_defaults.merge!(defaults) do |key, oldval, newval|
            if newval.respond_to?(:merge!) 
              newval.merge!(oldval)
            else
              newval
            end
          end
        end
        
        enable_const_missing()

        return all_defaults
      end

      # Given an array of filenames relative to the cookbooks, e.g. [ 'cookbooks/jke/recipes/create_database_mysql.rb' ]
      # generate a Hash containing the default values of the parameters
      def get_param_defaults(ruby_recipes, all_defaults)
        param_defaults = {}
        attempted_filenames = Hash[ruby_recipes.map { |filename| [filename, get_filename(filename)] }]
        attempted_filenames.each do |filename, ruby_recipe|
          if ruby_recipe.nil?
            @@logger.warn "Recipe #{filename} not found"
            @error_reporting.call(Logger::WARN, "Recipe #{filename} not found")
          end
        end
        
        attempted_filenames.values.select { |full_filename| full_filename }.each do |ruby_recipe|
          weaver_attr_names = scrape_param_usage(ruby_recipe)
          @@logger.debug "recipe #{ruby_recipe} uses parameters #{weaver_attr_names.inspect}"
          
          # each attribute name is a '.' separated list; the first entry is typically a cookbook name
          weaver_attr_names.each do |weaver_param_name|
            defaults_working_set = all_defaults
            attr_name_components = weaver_param_name.split(".")
            # @@logger.debug "evaluating attribute #{weaver_param_name}"

            key = attr_name_components.shift
            output_param_name = key.clone

            # loop over the components of the attribute name using them as keys to dig into 
            # all_defaults to find the corresponding default value
            until attr_name_components.empty?
              next_key = attr_name_components.shift

              # @@logger.debug "key = #{key}; next_key = #{next_key}"
              
              # lookup using key if can find desired next value
              modified_working_set = false
              if defaults_working_set.key?(key) && defaults_working_set[key].respond_to?(:fetch) # is a Hash      
                # @@logger.debug "dws[#{key}] is a Hash"          
                if contains?(defaults_working_set[key], next_key)
                  # @@logger.debug "dws[#{key}] contains #{next_key}"
                  defaults_working_set = defaults_working_set[key]
                  modified_working_set = true
                end
              end 
              
              # use symbol for key if didn't use key as a String
              defaults_working_set = defaults_working_set[key.to_sym] unless modified_working_set
              if defaults_working_set.nil?
                @@logger.warn "No defaults for #{key.to_sym.inspect}"
                defaults_working_set = {}
              end
              
              output_param_name << "." << next_key.to_s
              
              # we've hit the last component of the name; look up the final value
              if attr_name_components.empty? then
                if defaults_working_set.key?(next_key)
                  param_defaults[output_param_name] = defaults_working_set[next_key] unless defaults_working_set[next_key] == :multiple_values
                elsif defaults_working_set.key?(next_key.to_sym)
                  param_defaults[output_param_name] = defaults_working_set[next_key.to_sym] unless defaults_working_set[next_key.to_sym] == :multiple_values
                else
                  defaults_working_set[next_key] = nil
                  param_defaults[output_param_name] = nil
                end # if defaults_working_set.key?(next_key)
                # @@logger.debug %Q{assigned default #{output_param_name} => #{param_defaults[output_param_name]}}
              end
              
              key = next_key             
            end # end until attr_name_components.empty?
            
          end # weaver_attr_names.each do
        end # ruby_recipes.map do
        
        @@logger.debug "defaults calculated as #{param_defaults.inspect}"
        
        return param_defaults
      end # def get_param_defaults
      
      def contains? (working_set, key)
        # @@logger.debug "contains? called with key = #{key}; ws = #{working_set.inspect}"
        return false if working_set.nil? || working_set.empty? || key.nil?
        return true if working_set.key?(key)
        return true if working_set.key?(key.to_sym) 
        return false
      end
      
      def defaults_to_params(param_defaults, parent)
        param_defaults.each do |attr, val|
          property = Attribute.new(attr.symbolize, parent)
          is_pwd = should_encrypt?(property, val)
          property.value =  is_pwd ? (val.nil? || val.empty? ? val : EncryptedValue.from_plaintext(val, @keyfile || ENV['KEY_FILE'])) : val
          property.mutable = true
          property.required = false
          property.name = attr
          property.tags.add(:password) if is_pwd
          
          if val.kind_of?(String)
            property.constrain_value_type(:string)
          elsif val.kind_of?(Fixnum)
            property.constrain_value_type(:numeric)
          end
          
          if property.name =~ /\#{/
            @@logger.warn "Skipping calculated property name #{property.name}"
            @error_reporting.call(Logger::WARN, "Skipping calculated property name #{property.name}")
          elsif val.kind_of?(Hash)
            @@logger.warn "Skipping property name #{property.name} because #{val} is not a supported Weaver default"
            @error_reporting.call(Logger::WARN, "Skipping property name #{property.name} because #{val} is not a supported Weaver default")
          # The problem here is many defaults use Ruby interpolations to read Chef attributes, e.g.
          #   default['modjk']['loadfile'] = "#{node['apache']['dir']}/#{default['apache']['modules']['confdir']}/mod_jk.load"
          # Weaver doesn't know the value of node['apache']['dir'] and DefaultsProcessorRecursive incorrectly believes
          # node['apache']['dir'] is the empty Hash.  TODO instead of looking for {} which might be valid in a String
          # DefaultsProcessorRecursive should detect this problem and directly.
          elsif val.kind_of?(String) && val =~ /{}/
            @@logger.warn "Skipping intepolated property name #{property.name} because #{val} was not imported correctly"
            @error_reporting.call(Logger::WARN, "Skipping intepolated property name #{property.name} because #{val} was not imported correctly")
          else
             parent.amd[property.id] = property
          end
        end
      end

      def filter_parameters(hash_of_hashes, weaver_params)
        retval = hash_of_hashes.clone
        
        retval.each do |cookbook, parameters|
          parameters.delete_if { |parameter, value| !weaver_params.member?("#{cookbook}.#{parameter}") }
        end
        
        retval.delete_if { |cookbook, parameters| parameters.length == 0 }
          
        return retval
      end

      # Convert a \Chef default values hash into a chef_node_attr with \Weaver attributes for the values      
      def weaverize(hash_of_hashes)
        retval = {}

        # First, replace all default values with the Weaver property it refers to
        hash_of_hashes.each do |cookbook, parameters|
          newparams = {}
          parameters.each do |param, val|
            newparams[param] = UnquotedString.new("#{cookbook}.#{param}".symbolize.to_s)
          end
          
          retval[cookbook] = newparams
        end
        
        return retval
      end
      
      protected
      
      def multiparse_defaults(default_expr, default_file)
        vm_defaults = @@SIMULATED_VM_ATTRIBUTES.map { |h| parse_defaults(DefaultProcessorRecursive.new(h), 
          default_expr, default_file) }.select { |h| !h.nil? }
            
        # Compute the intersection of key-value pairs, returning a Hash of Hashes containing those pairs
        # in every element of vm_defaults
        accum = {}
        key_value_intersection(vm_defaults, accum)
        return accum
      end
      
      # Given an array of Hashes, fill an accumulator with the unique key-value pairs; if
      # they value isn't unique replace it with `:multiple_values`.
      # 
      # @param hashes [Hash[]] hashes
      # @param accum [Hash] out result
      # RECURSIVE
      def key_value_intersection(hashes, accum)
        return if hashes.empty?
        
        hashes.first.each do |key, val|
          if val.kind_of?(Hash)
            sub_accum = {}
            key_value_intersection(hashes.map { |h| h[key]}.select { |h| !h.nil? }, sub_accum)
            accum[key] = sub_accum unless sub_accum.empty?
          else
            possible_vals = hashes.map { |h| h[key] }.uniq
            accum[key] = (possible_vals.size == 1) ? possible_vals[0] : :multiple_values
          end
        end
      end

      # The defaults calculated by evaluating a Chef default attribute file under a particular VM configuration      
      def parse_defaults(default_processor, default_expr, default_file)
        begin
          default_processor.instance_eval(default_expr, default_file)
          return default_processor.default()
        rescue NoMethodError => e
          @error_reporting.call(Logger::WARN, "Couldn't parse default file #{default_file}; #{e.to_s}")
          @@logger.warn "#{e.to_s}\n#{e.backtrace.join("\n")}"
          return nil
        rescue StandardError => se
          @error_reporting.call(Logger::WARN, "Couldn't parse default file #{default_file}; #{se.to_s}")
          @@logger.warn "#{se.to_s}\n#{se.backtrace.join("\n")}"
          return nil
        end
      end
      
    end # class
    
    protected
    
    # mini-DSL to parse a role file
    class RoleProcessor
      def name(*val)
        return @name if val.empty?
        @name = val[0] 
      end

      # Parse Chef's +description+ statement 
      def description(*val)
        return @description if val.empty?
        @description = val[0]
      end

      # Parse Chef's +run_list+ statement 
      def run_list(*val)
        return @run_list if val.empty?
        @run_list = val
      end
      
      # Parse Chef's +override_attributes+ statement 
      def override_attributes(*val)
        return @override_attributes if val.empty?
        @override_attributes = val[0]
      end
      
      # Parse Chef's +default_attributes+ statement 
      def default_attributes(*val)
        return @default_attributes if val.empty?
        @default_attributes = val[0]
      end
      
    end
    
    # @private
    # A String that *doesn't* have quotes when inspected
    class UnquotedString < String # :nodoc:
      def initialize(s)
        super
      end
      
      def inspect
        return to_s
      end
    end
    
  end
end 