require 'weaver/cli/weaver_cli_plugin'
require 'weaver/chef/chef_node_inspector'
require 'devops_core/util'
require 'devops_library/util'
require 'devops_library/providers/ram_library_provider'
require 'open-uri'
require 'fileutils'

module Weaver
  module Cli
    module Plugins

      # Weaver CLI plugin for packaging in RAM (Rational Asset Manager)
      #
      # Syntax: weaver ram [name] [options]
      #
      # Author:: Waldemar Hummer
      #
      class RamCliPlugin < Weaver::Cli::WeaverCliPlugin

        namespace :ram
        namespace_description "Rational Asset Manager (RAM) related tasks"

        desc "package", "Package an archive file or directory to RAM (Rational Asset Manager)."
        method_option :descriptor, :required => true, :type => :string, :aliases => "-d", :desc => "RAM asset descriptor file (in JSON format)"
        method_option :content, :required => true, :type => :string, :aliases => "-c", :desc => "Package content (local directory or archive file)"
        method_option :properties, :required => true, :type => :string, :aliases => "-p", :desc => "YAML properties file for cloud deployer."
        method_option :name, :required => true, :type => :string, :aliases => "-n", :desc => "Name of the asset"
        method_option :version, :required => false, :type => :string, :aliases => "-v", :default => "1.0.0", :desc => "Version of the asset (default 1.0.0)"
        method_option :type, :required => false, :type => :string, :aliases => "-t", :default => "inf", :desc => "Type ('inf'=Infrastructure (default) or 'app'=Application)"
        method_option :bindings, :required => false, :type => :string, :aliases => "-b", :desc => "Bindings (e.g., APPLICATION_ID=... or INFRASTRUCTURE_VERSION=...)"
        method_option :overwrite, :required => false, :type => :string, :aliases => "-o", :desc => "Overwrite file if archive needs to be built and file already exists."
        method_option :remove, :required => false, :type => :boolean, :default => false, :desc => "Remove and re-create the whole RAM asset before uploading file contents (reloads asset dependencies)."
        #method_option :force, :required => false, :type => :boolean, :default => false, :aliases => "-f", :desc => "Force overwrite, even if existing files are identical."
        def package()
          collect_properties
          setup_logging

          content = options[:content]
          if content.end_with?("/")
            content = content[0..-2]
          end
          descriptor = options[:descriptor]

          # check if DESCRIPTOR_FILE parameter has been provided and exists
          say_status("Error", "DESCRIPTOR_FILE parameter needs to point to a file (provided: '#{descriptor}').", :red) && exit(1) unless File.file?(descriptor)

          # check if DIR_OR_FILE parameter has been provided and exists
          say_status("Error", ":content parameter needs to point to an archive file or directory.", :red) && exit(1) unless File.file?(content) || File.directory?(content)

          # check if :type parameter is either 'app' or 'inf'
          say_status("Error", ":type parameter needs to be either 'inf' (default) or 'app'.", :red) && exit(1) unless options[:type] == 'app' || options[:type] == 'inf'
          
          # check if :properties parameter is provided
          say_status("Error", ":properties parameter needs to point to a valid configuration file.", :red) && exit(1) unless options[:properties] && File.file?(options[:properties])

          bindings = options[:bindings]
          bindingsMap = {}
          if bindings
            bindings.split(/[; ]/).each { |str|
              parts = str.split("=")
              if parts.length == 2
                bindingsMap[parts[0].to_sym] = parts[1]
              end
            }
          end

          if options && options[:verbose]
            say_status("Parsing:", "Map of variable bindings: #{bindingsMap}", :green) unless bindingsMap.empty?
          end

          # If the provided path points to a directory, zip the contents of the directory into an archive..
          archive_file = content
          if File.directory?(content)

            archive_file = "#{content}.tgz"
            say_status("Error", "Archive file '#{archive_file}' exists and no overwrite option given.", :red) && exit(1) unless options[:overwrite] || !File.file?(archive_file)

            say_status("Info", "Packing contents of directory #{content} into archive #{archive_file}", :green)

            # The "GZIP='--no-name'" option ensures that the original file name is not encoded into the *.tgz file.
            # This is required because the filenames may change, but we rely on two archives with the
            # same *contents* to have the same *MD5 hash* (which requires that the files are identical byte-by-byte).
            # 
            # Note that it is *not* critical if the parameter or this mechanism fails (for whatever reasons). 
            # If it fails, we simply cannot take advantage of the performance improvement of comparing local and 
            # remote files to avoid redundant uploads, but the code itself remains functional.
            system "GZIP='--no-name' tar -zcvf #{archive_file} -C #{content} . > /dev/null"

            content = archive_file
          end

          props_file = options[:properties]
          props = YAML::load_file(props_file)

          template_file = descriptor

          # load asset config
          pkg_conf = DevopsLibrary::Util::load_library_asset_config(template_file, bindingsMap)
          
          # Setup the library gateway to package and publish the application to Rational Asset Manager
          use_mock = false
          asset_library = DevopsLibrary::Providers::RAMLibraryProvider.newFor(
            use_mock,
            props["app_lib_properties"]
          )

          if options[:remove]
            say_status("Info", "Completely removing the entire asset '#{options[:name]}[#{options[:version]}]' from RAM.", :green)
            asset_library.delete_package(options[:name], options[:version])
          end

          filename = options[:name] + ".tgz"

          # check if the file(s) already exists in RAM. if so, compare MD5 hash of file(s)
          repo_url = props["app_lib_properties"]["repository_url"]
          file_url = "#{repo_url}/oslc/assets/#{options[:name]}/#{options[:version]}/artifactContents/#{filename}"

          #require tmpdir
          file_exists = false
          mktmpdir {|tmpdir|
            begin
              existing_pkg = asset_library.resolve_package(options[:name], options[:version], {})
              if existing_pkg 
                say_status("Downloading:", "Saving #{file_url} to temporary directory to compare file hashes.", :green)
                asset_library.fetch_package_file(options[:name], options[:version], filename, tmpdir)
                require 'digest/md5'
                digest1 = Digest::MD5.hexdigest(File.read("#{tmpdir}/#{filename}"))
                digest2 = Digest::MD5.hexdigest(File.read(archive_file))
                file_exists = (digest1 == digest2)
                say_status("Info", "Archive file hash is different, overwriting file.", :green) unless file_exists
              end
            rescue => ex
              # swallow
            end
          }

          say_status("Info", "Archive file #{filename} is already published (MD5 file hashes have been compared) and no overwrite option given.", :green) && return unless !file_exists || options[:overwrite]

          say_status("Loaded:", "RAM asset configuration file #{template_file}", :green) if options && options[:verbose]

          # Create an asset in the asset library for the package
          asset = asset_library.create_package(
            options[:name], 
            options[:version], 
            pkg_conf)

          if options[:type] == 'inf'
            # Publish BOTH the infrastructure file and the deps file to the same asset in the asset library
            infrastructure_url = asset_library.publish_package(asset, content, filename)
            say_status("Info", "Published to RAM: #{infrastructure_url}", :green)

            deps_url = asset_library.publish_dependency_file(asset)
            say_status("Info", "Published to RAM (dependency file): #{deps_url}", :green)

            #puts "You may now deploy your application using the following command (or similar):"
            #puts "weaver env deploy -c iwd -p deploy_properties.yml -d \"deps_file_url=#{deps_url}\" environment_file.weaver"
          elsif options[:type] == 'app'

            # Publish the application file
            app_url = asset_library.publish_package(asset, content, filename)
            say_status("Info", "Published to RAM: #{app_url}", :green)

          end
        end

        no_tasks do
          # Send a message to the user
          def report_message(status, level, msg)
            if level == Logger::ERROR || level == Logger::FATAL
              say_status(status, msg, :red)
            elsif level == Logger::WARN
              say_status(status, msg, :yellow)
            else
              say_status(status, msg, :green)
            end
          end
        end
        
        private 

        # Dir.mktmpdir fails under certain JRuby versions, hence we 
        # provide our own simple implementation here..
        # see, e.g., http://jira.codehaus.org/browse/JRUBY-6178
        def mktmpdir(&block)
          file = Tempfile.new('weaver_tmp_dir')
          path = file.path
          file.unlink
          Dir.mkdir(path)
          begin
            block.call(path)
          ensure
            FileUtils.rm_rf(path)
          end
        end

      end
    end

  end
end