require 'weaver/core/base.rb'

module Weaver
  module Core
    # A mixin for classes such as Application which may contain Group
    # 
    # Author:: Ed Snible
    # Copyright:: IBM
    module ImportContainerMixin
      include Weaver::Core
      
      # Let all classes that include ImportContainerMixin have a method to return their weaver_imports
      def self.included(baseclazz)
        baseclazz.class_exec {
          # Hash of id -> Import
          def weaver_imports
            # Since mixins have no initialize() to set up @weaver_imports, initialize @weaver_imports lazily
            @weaver_imports ||= {} unless frozen?
            return @weaver_imports || {}
          end
        }
        
        baseclazz.register_validate_handler(:validate_imports)
      end
    
      # Represents importing a .weaver file into this top-level weaver document for refinement
      # * *Args*    :
      #   - +namespace+ -> a Symbol representing the namespace, or nil
      #   - +import+ -> a Weaver::Core::Import
      #
      def import(namespace, import, filespec)
        # Save the filenames the imports arrived in
        @raw_weaver_imports = Hash.new { |ns, files| ns[files] = [] } if @raw_weaver_imports.nil?
        @raw_weaver_imports[namespace].push(filespec)
        
        if weaver_imports.key?(namespace)
          @@logger.debug "Merging imports in namespace #{namespace.inspect}"
          orig = weaver_imports[namespace]
          
          # Merge topologies
          import.topologies.each do |id, top|
             dup = Weaver::Transforms::Duplicate.duplicate(top)
             dup.scope = orig
             orig.topologies[id] = dup
             # TODO handle ID collisions
          end
          
          # Merge load_drivers
          import.load_drivers.each do |id, ld|
             dup = Weaver::Transforms::Duplicate.duplicate(ld)
             dup.scope = orig
             orig.load_drivers[id] = dup
             # TODO handle ID collisions
          end
          
          # Merge policies
          import.policies.each do |id, policy|
             dup = Weaver::Transforms::Duplicate.duplicate(policy)
             dup.scope = orig
             orig.policies[id] = dup
             # TODO handle ID collisions
          end
          
          # Merge nodes
          import.nodes.each do |id, node|
            dup = Weaver::Transforms::Duplicate.duplicate(node)
            dup.scope = orig
            orig.nodes[id] = dup
            # TODO handle ID collisions
          end
          
          # Merge services
          import.services.each do |id, service|
            dup = Weaver::Transforms::Duplicate.duplicate(service)
            dup.scope = orig
            orig.services[id] = dup
            # TODO handle ID collisions
          end
          
          @@logger.debug "Namespace #{namespace.inspect} now contains #{orig}"
          
          # TODO warn or merge if import has other types (Import, etc)
          return orig
        end
        
        weaver_imports[namespace] = import
      end
      
      # Validate the imports
      # * +out_messages+ ProblemStatus will be pushed onto this for any validation message
      # * +desired_phases+ an optional Array which may be +[:predeployment]+
      def validate_imports(out_messages, desired_phases = nil)
        weaver_imports.each_value do |imp|
          # Imports can be nested, but do not appear in child_entities() or flatten_children() so check them explicitly
          # looking for validation problems with indirectly imported source files.
          imp.validate_imports(out_messages, desired_phases)
          
          # We can't merely call imp.validate_collect because the original imported entity never got
          # augmented and may have many not_empty() etc. validators failing.  Just look for :always_fail
          # validators such as file not found and syntax error 
          [imp].concat(imp.flatten_children).each do |imported_entity| 
            imported_entity.validators.each do |symb, *params|
              if symb == :always_fail
#                imported_entity.call_validator(symb, params, out_messages, desired_phases)
                report(Logger::ERROR, "#{params[0]} at #{params[1]}", imp, nil, nil, nil, [], nil)
              end
            end
          end
        end
      end
      
    end
  end
end