require 'weaver/transforms/unroll'

module Weaver
  module Core

    # This class represents a Weaver +late_binding+.  LateBinding holds a Ruby Proc allowing procedural code to
    # be used within a Weaver module.  Attributes may be assigned a LateBinding as a value.  Such attributes value
    # changes, usually when other attribute values change.   
    #
    # Example
    #    pattern.name = late_binding { "DevOps_Money_That_Matters_Weaver_" + build_tag }  
    #
    # Author:: Ed Snible
    class LateBinding
      
      # Proc to evaluate the binding
      attr_reader :proc
      
      # Arguments to pass to the proc (usually [])
      attr_reader :extra_args 
      
      # The scope this late binding was declared upon
      attr_reader :declaring_scope  
      
      # +expr+ is the expression that makes the late binding; something that can appear inside #{}
      # +declaring_scope+ is the entity the binding should be evaluated upon
      # +extra_args+ allows the late binding to receive arguments.  EXPERIMENTAL
      # +block+ the block to evaluted as needed
      def initialize(expr, declaring_scope, extra_args = [], &block)
        @expr = expr
        @declaring_scope = declaring_scope
        raise WeaverError.new("declaring_scope must be a Base") unless declaring_scope.kind_of?(Base)
        @extra_args = extra_args
        @proc = block
      end
      
      # Evaluate a Late Binding.  The #proc is evaluated within a Sandbox.  The Sandbox works in cahoots with
      # LateBinding to produce a Set of Attributes which were consulted during the evaluation.  This Set can be
      # used for propagation detection.
      #
      # * +entity+ the scope of the evaluation which is used for looking up symbols
      # * +out_dependencies+ a Hash, which upon return will contain the Attribute s used to evaluate the binding
      def weaver_eval(entity, out_dependencies = {}, orig_dependencies = {})
        # Note that we are now ignoring +entity+, because @declaring_scope is what we should evaluate using
        sandbox = Sandbox.new(@declaring_scope, entity, '', out_dependencies, orig_dependencies, AllCache.new(out_dependencies))
        retval = sandbox.instance_exec(extra_args, &@proc)
        
        return retval
      end
      
      # Returns +true+ for any other LateBinding
      def ==(another_binding)
        return false unless another_binding.kind_of?(LateBinding)
        
        # Don't even attempt to peek into LateBinding for equality checking as it would
        # involve evaluating proc
        return true
      end
      
      # If this late binding was created from an String containing an original expression tries to create Weaver code that will recreate
      # this LateBinding.  If this binding was only created from a Proc just creates a bogus late binding that
      # the user will need to improve by hand.
      def inspect
        return "late_binding { #{@expr || ':ruby_code' } }"
      end
      
      # Returns an Array of Attributes this LateBinding uses upon to reach its value
      def attribute_dependencies(entity)
        out_dependencies = {}
        sandbox = Sandbox.new(@declaring_scope, entity, '', out_dependencies, {}, AllCache.new(out_dependencies))
        bindings_value = sandbox.instance_exec(@extra_args, &@proc)
        # It isn't enough to instance_eval parse it, we must also trigger eval
        bindings_value.to_s
        return out_dependencies
      end
      
      # Creates a new Late Binding; the new binding's #weaver_eval method should return a concatenation.
      #
      # Currently only implemented for concatentating Strings, not another late binding.
      def +(obj)
        if obj.kind_of?(String)
          # Late Bindings are evaluated using SandBox.instance_eval()
          # TODO escape 'obj' so that it can safely appear in single quotes.
          code = "Proc.new { |late_binding| late_binding.weaver_eval(@entity, @dependencies) + #{obj.dump} }"
          return LateBinding.new(nil, declaring_scope, self, &(eval(code)))
        end
        
        raise WeaverError.new("Unimplemented: concatenate late_binding with #{obj.class}")
      end
      
      # Output something that looks like the late binding
      def to_s
        return '#{' + @expr + '}' unless @expr.nil?
        super
      end
      
      # Returns the source of this LateBinding (if possible) using the sourcify gem.  
      # Format: late_binding { ... }    
      def to_source (options = {})
        if @expr.nil?
          # we don't know the ruby code
          begin
            # Suppress ruby_parser complaints exposed by sourcify
            orig_stdout = $stdout
            $stdout = StringIO.new
            return @proc.to_source(options).gsub('proc', 'late_binding')
          ensure
            $stdout = orig_stdout
          end
        else
          # we do know the ruby code
          return 'late_binding {' + @expr + '}' unless options[:strip_enclosure]
          return @expr # if [:strip_enclosure]
        end # if @expr.nil?
      end
      
      # Repoint scope(s) after deep copy or fold
      # +old_to_new+ Hash of Base -> Base, called after clone or Fold::fold
      def freshen_declaring_scope(old_to_new)
        # Freshen if we can.  Note that during folding we might not need to update.  Note that multiple
        # Attributes can point to the same LateBinding so update may be requested twice for the same
        # unresolved_value.  Thus we can't insist on updating every time we reach this method.
        if old_to_new.key?(declaring_scope)
          dup = old_to_new[declaring_scope]
          @declaring_scope = dup
        end

        # We will have nested late bindings sometimes, such as after concatenating a String with a LateBinding        
        if @extra_args.kind_of?(LateBinding)
          @extra_args.freshen_declaring_scope(old_to_new)
        end
      end
    end
 
    # This class implements a 'sandbox' for in which we evaluate late_binding expressions.
    private 
    class Sandbox
      
      # *Args*
      #  - +weaver_entity+ the entity being evaluated
      #  - +execution_context+ the target of the value being assigned
      #  - +entity_ref+ the (partial) text reference to the weaver_entity
      #  - +out_dependencies+ a list of the attributes found during evaluation
      #  - +out_values+
      #  - +all+ flag indicating that the evaluation of an attribute should be to return all the values of the attribute from all members of the group to which the node is a member.  The order is not specified.
#      def initialize(weaver_entity, execution_context, entity_ref, out_dependencies, parent_sandbox, all_cache)
      def initialize(weaver_entity, execution_context, entity_ref, out_dependencies, parents_dependencies, all_cache)
        raise WeaverError.new("'weaver_entity' must be a Base but it is a #{weaver_entity.class}, caller=#{caller}") unless weaver_entity.kind_of?(Base)
        raise WeaverError.new("'out_dependencies' must be a Hash; not a #{out_dependencies.class}") unless out_dependencies.kind_of?(Hash)
        @entity = weaver_entity
        @context = execution_context
        @dependencies = out_dependencies
        # Keep track of the reference as it appears in the late binding expression
        @entity_ref = entity_ref
        # Remember the Weaver::Core::Node associated with +execution_context+
        @node_context = @context.scope_of_type(Weaver::Core::Node) unless @context.nil?
        # Keep the track of the "parents" dependencies so we can avoid problems when evaluating x + x
        @parents_dependencies = parents_dependencies
        # Keep track of computed values for expr and for all(expr)
        @all_cache = all_cache

        # For debugging        
        @sbid = (0...8).map{(65+rand(26)).chr}.join
#        puts "#{@sbid}: Sanbox created; grand-parent = #{@parent.id}" unless @parent.nil?
#        puts "#{@sbid}: Sanbox created; entity = #{@entity.fully_qualified_id} (#{@entity_ref})" unless @entity.nil?
#        puts "#{@sbid}: Sanbox created; dependencies = #{Sandbox.log_dependencies(@dependencies)}"
      end
      
      def attribute_dependencies
        # Merge any child dependencies
        return @dependencies
      end
      
      protected
      
      # For debugging
      def id
        return @sbid
      end # def sbid
      
      # Find the container of type clazz and return it and the path from the container to the input entity
      # *Args*
      #  - +entity+ the entity whose parent is sought
      #  - +clazz+ the type of the container sought
      # *Returns*
      #  - +entity+ the container of the input entity
      #  - +id+ the path (dotted) from the container to the input entity
      # Follow 'scope' until we encounter 'clazz', return nil if not found
      def container_path(entity, clazz)
        id = nil
        while !entity.nil? && !entity.kind_of?(clazz)
          id = (id.nil? or id.empty?) ? "#{entity.id}" : "#{entity.id}.#{id}"
          entity = entity.scope
        end
        return entity, id
      end

      # add a new dependencies ot a list (if it isn't already there)
      # dependencies is an out attribute     
      # return true if attr_ref is new 
      def add_dependency(attr, attr_ref, dependencies)
        attribute_references = dependencies[attr] || Set.new
        already = attribute_references.member?(attr_ref)
        unless already
          attribute_references.add(attr_ref)
          dependencies[attr] = attribute_references
        end # unless already
        return !already
      end # def add_dependency
      
      # Remove an entry from a Hash of dependencies.  If the mapped Set is empty, set to nil.
      def remove_dependency(attr, attr_ref, dependencies)
        refs = dependencies[attr]
        unless dependencies[attr].nil?
          dependencies[attr].delete(attr_ref)
          dependencies.delete(attr) if dependencies[attr].empty?
        end # unless dependencies.nil?
      end # def remove_dependency
      
      def evaluate_attribute (attr, attr_ref, dependencies)
#        puts "#{@sbid}: evaluate_attribute called with #{attr} (#{attr_ref})" 
        unless attr.unresolved_value.kind_of?(LateBinding)
          add_dependency(attr, attr_ref, dependencies)
          retval = attr.value
#          puts "#{@sbid}: evaluate_attribute returning (a) #{retval}"
          return retval 
        end # unless attr.unresolved_value.kind_of?(LateBinding)
        
        if @parents_dependencies.member?(attr)
#          puts "#{@sbid}: evaluate_attribute: returning recusive reference"
#          puts "#{@sbid} parents_dependencies are\n#{Sandbox.log_dependencies(@parents_dependencies)}"
          return :recursive_reference
        end
                      
        add_dependency(attr, attr_ref, dependencies)
        # pass copy of parents dependencies for recusive checking but add the new one
        updated_parents_dependencies = @parents_dependencies.clone
        add_dependency(attr, attr_ref, updated_parents_dependencies)
        attr_value = attr.unresolved_value.weaver_eval(@entity, dependencies, updated_parents_dependencies)
#        puts "returning (c) #{attr_value}"
        return attr_value
      end # def evaluate_attribute
      
      def evaluate_attributes (attr, attr_ref)
#        puts "#{@sbid}: Found an attribute, evaluating #{attr.fully_qualified_id}"
#        puts "#{@sbid}: calling evaluate_attributes for default value"
        @all_cache.value = evaluate_attribute(attr, attr_ref, @all_cache.dependencies)
#        puts "#{@sbid}: evaluate_attributes cached value #{@all_cache.value} with dependencies\n#{Sandbox.log_dependencies(@all_cache.dependencies)}"
        node, q_id = container_path(attr, Weaver::Core::Node)
        unless node.nil?
#          puts "Scope lookup found node = '#{node.fully_qualified_id}' with attribute qualifier '#{q_id}'"
          group = Weaver::Transforms::Unroll.group(node)
          # group = Weaver::Transforms::Unroll.template?(node) ? Weaver::Transforms::Unroll.group(node) : nil
#          puts "node group is #{group.fully_qualified_id}" unless group.nil?
          nodes = group.nil? ? [node] : group.members.values.select {|n| !Weaver::Transforms::Unroll.template?(n)}
#          puts "nodes are #{nodes.map {|n| n.fully_qualified_id}.join(', ')}"
          recursive = false
          @all_cache.all_values = nodes.map do |node|
            prefix = q_id.match(/(.*)\.[^\.]*$/)
            prefix = prefix[1] unless prefix.nil?
            suffix = q_id.match(/[^\.]*$/)[0]
            base = prefix.nil? ? node : node.instance_eval(prefix)
            attr = base.amd[suffix.to_sym]
#            puts "#{@sbid}: calling evaluate_attributes for array entry"
            value = evaluate_attribute(attr, attr_ref, @all_cache.all_dependencies)
#            puts "#{@sbid}: array entry value is #{value}"
            if value == :recursive_reference
              recursive = true
#              puts "#{@sbid}: breaking out of each loop"
              break
            end
            value
          end # node.map do
          @all_cache.all_values = :recursive_reference if recursive
        end # unless node.nil?
#        puts "evaluate_attributes precomputed values are: #{@all_cache.all_values}"
          
        # Return value is assumed to be a single reference; if wrapped by all() will deal with later
        # Update @dependencies to include the dependencies computed when computing the single reference
        # But... keep track of which were newly added in case an all() exists ... in which case we will need to undo the addition
#        puts "#{@sbid}: Plan to return singleton value #{@all_cache.value} for #{attr_ref}"
#        puts "#{@sbid}: Initial dependencies are:\n#{Sandbox.log_dependencies(@dependencies)}"
#        puts "#{@sbid}: evaluate_attributes; Plan to add:\n#{Sandbox.log_dependencies(@all_cache.dependencies)}"
        @all_cache.dependencies.each do |attr, refs|
          refs.each { |ref| add_dependency(attr, ref, @all_cache.new_dependencies) if add_dependency(attr, ref, @dependencies) }
        end # @all_cache.dependencies.each do
        @all_cache.dependencies.clear
#        puts "#{@sbid}: Final dependencies are:\n#{Sandbox.log_dependencies(@dependencies)}"
#        puts "#{@sbid}: Parent's original dependencies are:\n#{Sandbox.log_dependencies(@parents_dependencies)}"
#        puts "#{@sbid}: Also recorded as new: \n#{Sandbox.log_dependencies(@dependencies)}"
        
#        puts "#{@sbid}: Cached values (array) is #{@all_cache.all_values}"
#        puts "#{@sbid}: Cached dependencies are:\n #{Sandbox.log_dependencies(@all_cache.all_dependencies)}"
        ret_value = @all_cache.value
        return ret_value
      end # def evaluate_attributes

      # all() called on an attribute. Instead of returning the current single value (arg), return values 
      # from each group member as an array. Recall that these values were precomputed and stored in 
      # \@all_cache.all_values
      def all (arg)
#        puts "#{@sbid}: all called. arg (current value) is #{arg}"
        # We had earlier (cf. end of evaluate_attributes()) added @all_cache.dependencies, those dependencies
        # that made sense for the single value initially returned (@all_cache.value).
        # However, now that all() as been called, we want to remove those dependencies (any new ones) and 
        # replace them with those defined for a set of values; ie, @all_cache.all_dependencies
        # Finally, return the set of precomuted values as an array (@all_cache.all_values)
#        puts "#{@sbid}: Start removal of dependencies"
#        puts "#{@sbid}: Initial dependencies are:\n#{Sandbox.log_dependencies(@dependencies)}"
#        puts "#{@sbid}: Plan to remove:\n#{Sandbox.log_dependencies(@all_cache.new_dependencies)}"
        @all_cache.new_dependencies.each { |attr, refs| refs.each { |ref| remove_dependency(attr, ref, @dependencies) } }
        @all_cache.new_dependencies.clear
#        puts "#{@sbid}: Completed removal of dependencies; add all dependencies"
#        puts "#{@sbid}: Initial dependencies are:\n#{Sandbox.log_dependencies(@dependencies)}"
#        puts "#{@sbid}: all; Plan to add:\n#{Sandbox.log_dependencies(@all_cache.all_dependencies)}"
        @all_cache.all_dependencies.each { |attr, refs| refs.each { |ref| add_dependency(attr, ref, @dependencies) } }
        @all_cache.all_dependencies.clear
#        puts "#{@sbid}: Final dependencies are:\n#{Sandbox.log_dependencies(@dependencies)}"
#        puts "#{@sbid}: Parent's original dependencies are:\n#{Sandbox.log_dependencies(@parents_dependencies)}" 
#        puts "#{@sbid}: all; returning #{@all_cache.all_values}"
        return @all_cache.all_values.clone
      end # def all
      
      def method_missing(sym, *args, &block)
#        puts "#{@sbid}: method_missing called with symbol #{sym.inspect}"
        # Does entity have a corresponding Ruby method?  Ruby methods do not have propagation analysis
        if @entity.original_respond_to?(sym)
#          puts "#{@sbid}: Using corresponding ruby method"
          return @entity.send(sym, *args)
        end
        
        # Keep track of the reference to the entity; this will be returned to the caller as part of the structure of dependencies
        # An entry in out_dependencies is a Hash entry mapping the attribute to its entity_ref 
        updated_entity_ref = "#{@entity_ref}"
        updated_entity_ref << '.' unless updated_entity_ref.empty?
        updated_entity_ref << sym.to_s
#        puts "#{@sbid}: updated_entity_ref computed as #{updated_entity_ref}"
        
        # Treat "x" as a synonym for "amd[:x].value"
        attr = @entity.amd[sym]
        if !attr.nil?
          return evaluate_attributes(attr, updated_entity_ref)
        end
      
        # Treat "x" as a synonym for "scoped_lookup(:x)"
        entity = @entity.scoped_lookup(sym)
#        puts "#{@sbid}: Did scoped_lookup ==> entity: #{entity.fully_qualified_id}"
        
        # If we got an Attribute we don't expect a '.' and an attribute after it, so assume we should evaluate
        if entity.kind_of?(Attribute)
          attr = entity
          return evaluate_attributes(attr, updated_entity_ref)
        end
        
        # We got a (non-Attribute) entity, return a sandbox to be the left-hand-size of a <expression>.attribute
        # If the entity is a node, then check the calling context.  If the context was a group member and the entity
        # is the template node for the group, reset entity to be the group member (that was the calling context).
        # We do this to make sure we return the value of the attribute on the member instead of on the template.
        if entity.is_a?(Weaver::Core::Node)
          if Weaver::Transforms::Unroll.template?(entity)
            unless @node_context.nil?
#              puts "#{@sbid}: entity (#{entity.fully_qualified_id}) is template; node_context is (#{@node_context.fully_qualified_id})"
              if Weaver::Transforms::Unroll.group(@node_context)  == Weaver::Transforms::Unroll.group(entity)
                entity = @node_context
              end # if same group
            end # if !@node_context.nil? then
          end # if entity is a template
        end # if entity.is_a?(Weaver::Core::Node)
        if !entity.nil?
#          puts "#{@sbid}: creating sandbox with new entity (#{entity})"
          child_sandbox = Sandbox.new(entity, @context, updated_entity_ref, @dependencies, @parents_dependencies, @all_cache)
          return child_sandbox
        end
        
        raise WeaverError.new("Unknown #{sym.inspect} on :#{@entity.id}")         
      end # def method_missing?
      
      def self.log_dependencies (dependencies)
        dependencies.map { |attr, refs| "   #{attr.fully_qualified_id} (#{refs.to_a.join})" }.join("\n")
      end
      
    end # class LateBinding
    
    class AllCache
      attr_accessor :all_values
      attr_accessor :all_dependencies
      attr_accessor :value
      attr_accessor :dependencies
      attr_accessor :new_dependencies
      
      def initialize (dependencies = {})
        @all_values = []
        @value = nil
        @new_dependencies = {}
        @dependencies = Hash.new { |h,k| h[k] = Set.new }
        @all_dependencies = Hash.new { |h,k| h[k] = Set.new }
        dependencies.each do |attr, refs|
          refs.each do |ref|
            @dependencies[attr].add(ref)
            @all_dependencies[attr].add(ref)
          end # refs.each do |ref|
        end # dependencies.each do
      end # def initialize

    end # class
    
  end # module Core
end # module Weaver