require 'weaver/core/base'
require 'weaver/core/component'
require 'weaver/core/mixin'
require 'weaver/core/node_container_mixin'
require 'weaver/core/topology_container_mixin'
require 'weaver/core/link_container_mixin'
require 'weaver/core/group_container_mixin'
require 'weaver/core/user_container_mixin'
require 'weaver/core/provider_container_mixin'
require 'weaver/core/import_container_mixin'
require 'weaver/core/load_driver_container_mixin'
require 'weaver/core/policy_container_mixin'
require 'weaver/core/artifact_container_mixin'
require 'weaver/core/network_interface_controller'
require 'weaver/validation/validation_error'
require 'weaver/iwd/pattern_utils'

module Weaver
  module Core
    # represents a .weaver top-level (not included) file
    class Root < Base
      # Validation messages appear sorted in roughly the mixin include order, so important problems that can trigger others, 
      # like import failure, should be placed higher in this list.
      include ImportContainerMixin
      include ProviderContainerMixin
      include ComponentContainerMixin
      include AutomationContainer
      include NodeContainerMixin
      include LinkContainerMixin
      include LoadDriverContainerMixin
      include PolicyContainerMixin
      include GroupContainerMixin
      include UserContainerMixin
      include ServiceContainerMixin
      include TopologyContainerMixin
      include Weaver::Core::Generation
      include ArtifactContainerMixin
      
      # A String which will be serialized
      attr_accessor :preamble
      
      # Hash of ID to Rule
      # All rule {} statements encountered during parsing are here not only top-level ones.
      attr_accessor :rules
      
      # Hash of ID to Proc
      attr_reader :named_validators
      
      # Hash of ID to Base
      attr_reader :weaver_extended_types
      
      # :stopdoc:
      # Holds messages on behalf of validator.  Deprecated
      attr_reader :unreported_messages
      # :startdoc:
      
      # The Weaver::Environment that carries this model's configuration
      attr_accessor :environment
            
      # * +id+ A String or Symbol for the Root, typically the filename
      # * +parent+ The enclosing Base
      def initialize(id, parent = nil)
        super(id, nil)
        @type = :root
        @rules = {}
        
        # These are the base types from which we can create Extended types
        @weaver_extended_types = {
          :component => Component.new(:component, nil),
          :user => User.new(:user, nil),
          :group => Group.new(:group, nil),
          :link => Link.new(:link, nil),
          :config_link => Link.new(:config_link, nil),
          :load_driver => LoadDriver.new(:load_driver, nil),
          :policy => Policy.new(:policy, nil),
          :topology => Topology.new(:topology, nil),
          :automation => Automation.new(:automation, nil),
          :nic => NetworkInterfaceController.new(:nic, nil),
          :provider => Provider.new(:provider, nil),
          :mixin => Mixin.new(:mixin, nil),
          :service => Service.new(:service, nil),
          :base => Base.new(:base, nil),
          :artifact => Artifact.new(:artifact, nil)
        }
        
        # (We make this separately so we can refer to @weaver_extended_types[:nic] in Node's constructor)
        @weaver_extended_types[:node] = Node.new(:node, nil)
          
        @weaver_extended_types.each_value do |val|
          val.scope = self 
          val.freeze  # Prevent base prototypes from being modified
        end
        
        # We initialize the :always_fail validator here and not in validators.weaver because we want
        # it available immediately in case there are errors in validators.weaver
        @named_validators = { :always_fail => Validator.new(:always_fail, :any, Proc.new do |msg, origin|
          report(Logger::ERROR, msg, self, nil, nil, nil, [], origin) 
        end
        ) }
          
        # TODO Is there a way to push these validation messages through the stack?
        @unreported_messages = []
#        def unreported_messages.push(obj)
#          raise "Can't push #{obj.class}" unless obj.kind_of?(Weaver::Validation::ProblemStatus)
#          super
#        end
      end
      
      def initialize_copy(other)
        super
        
        # At this point the copy is shallow; make it deep
        @rules = other.rules.clone
        @named_validators = other.named_validators.clone
        @weaver_extended_types = @weaver_extended_types.clone
        @preamble = other.preamble
      end
      
      def validate_collect(out_messages, desired_phases = nil, excluded_tags = nil)
        @@logger.debug "Validating root '#{id}'"
        
        # We don't want to validate the prototype as an instance because we don't want to run things
        # like not_empty validators.  However if there were parsing errors we want to display them.
        @weaver_extended_types.each_value do |prototype|
          prototype.validators.each do |symb, *params|
            if symb == :always_fail
              prototype.call_validator(symb, params, out_messages, desired_phases)
            end
          end
        end
        
        out_messages.concat(unreported_messages)
        unreported_messages.clear
        
        super
        
        rules.each do |id, rule|
          rule.validate(self, out_messages, desired_phases, excluded_tags)
        end
      end
      
      # Note that Root gathers reports for all contained Weaver Base objects, not just itself
      # * *Args*    :
      #   - +level+ -> a Logger level; Logger::ERROR is usual
      #   - +message+ -> an untranslated message
      #   - +context+ -> a Weaver entity associated with the message
      #   - +type+ -> optional special type (e.g. an IWD problem)
      #   - +description+ -> optional additional details
      #   - +resolutions+ -> optional array of Resolutions with evaluated strings
      #   - +origin+ -> a file:lineno String if different than the line declaring 'context'
      def report(level, message, context, exception = nil, type=nil, description=nil, resolutions=[], origin=nil)
        raise WeaverError.new("'context' should be a Base but it is a #{context.class} at #{caller}") unless context.nil? || context.kind_of?(Base)
        problem_status = Weaver::Validation::ProblemStatus.new(level, message, context, origin, type, exception, description, Weaver::Iwd::PatternUtils.evaluated_resolutions(resolutions))
        @unreported_messages.push(problem_status)
        return problem_status
      end
      
      # Summarize model
      def to_s
        return "Root topologies #{topologies.inspect}, #{rules.size} rules, #{@weaver_extended_types.size} types"
      end
      
      # Return the controlling Provider.  This may be inherited.  If there are multiple ones
      # the provisioner is ambiguous but we will return one anyway, leaving any detection
      # of the problem to vaidation.
      def provisioner
        return providers.empty? ? nil : providers.values.to_a[0]
      end
      
      # Returns a prototype (subtype, subcomponent, created through the 'type' keyword)
      def prototype(clazzname)
        retval = @weaver_extended_types[clazzname]
        raise WeaverError.new("No base type #{clazzname}") unless retval
        return retval
      end
      
      # Create an instance of a prototype
      # * *Args*    :
      #   - +clazzname+ -> Weaver class, such as +:node+
      #   - +id+ -> ID to give the new instance
      #   - +scope+ -> parent to give the new instance
      #   - +origin+ -> a String, filename:lineno, to use when validating the instance
      def prototype_instance(clazzname, id, scope, origin)
        prot = @weaver_extended_types[clazzname]
        raise WeaverError.new("No base type '#{clazzname}' for instance") unless prot
        retval = Weaver::Transforms::Duplicate.duplicate(prot)  # We can't merely dup, because we might have inner objects
        retval.provenance = prot
        
        unless id.kind_of?(Symbol) || id.nil? || id.kind_of?(String)
          retval.add_validation_failure("Cannot create #{retval.type} with ID #{id} (a #{id.class})")
          id = :error
        end
        retval.id = id
        retval.scope = scope
        retval.set_origin(origin, true)
        return retval
      end
      
      # Make the entity a kind of new_prototype_instance by replacing the @type and adding implicit parameters
      def widen(entity, new_prototype_instance)
        entity.type = new_prototype_instance.type
        new_prototype_instance.amd.each do |id, attr|
          if attr.implicit? && entity.amd[id].nil?
            dup = attr.dup()
            dup.scope = entity
            dup.implicit = true
            entity.amd[id] = dup 
          end
        end
      end

      # Create a new Weaver class that can be re-used in Weaver documents during this run.
      def make_new_class(prototype)
        raise WeaverError.new("Type :#{prototype.id} cannot be redefined; original was defined by #{@weaver_extended_types[prototype.id].weaver_origin}") if @weaver_extended_types.key?(prototype.id)
        prototype.scope = self
        
        # Mark the attributes as 'implicit' meaning that they can take values the first time they are used
        prototype.flatten_attributes.each { |attr| attr.implicit = true }
          
        @weaver_extended_types[prototype.id] = prototype
      end
      
      # Can we handle a subcomponent by that name?
      def weaver_extended_type?(sym, clazz)
        prototype = @weaver_extended_types[sym]
        return prototype.kind_of?(clazz)
      end

      # Returns an Import imported under a symbol
      # * +sym+ a Symbol
      def namespace(sym)
        return weaver_imports[sym]
      end
      
      # Support for old test cases that assume root has a single app named 'app'
      def app
        return topologies.values.find { |top| top.weaver_instance_of?(:application) || top.cloud == nil }
      end
            
      # Support for old test cases that assume root has a single infrastructure named 'infrastructure'
      def infrastructure
        return topologies.values.find { |top| top.weaver_instance_of?(:infrastructure) || top.cloud != nil }
      end
            
      # Support for old test cases that assume root has a single infrastructure named 'infrastructure'
      def env
        # If there is only one top-level Topology and it contains child topologies it is the environment
        if topologies.size == 1
          candidate = topologies.values.to_a[0] 
          return candidate if candidate.topologies.size > 0
        end
        
        # If there are more pick the one with type=:environment
        return topologies.values.find { |top| top.weaver_instance_of?(:environment) }
      end

      # Support for old samples that use symbols for realization            
      def node_by_id(id)
        raise WeaverError.new("'id' must by a Symbol but it is a #{id.class}") unless id.kind_of?(Symbol)
        retval = flatten_children.find { |n| n.kind_of?(Node) && n.id == id }
        return retval.nil? ? id : retval
      end
      
      # Create an instance of a Provider
      #   - +provider_id+ -> Kind of provider (e.g. :amazon_ec2, :iwd_vsp)
      #   - +parent+ -> Core::Base
      def provider_instance(provider_id, parent)
        return Provider.new(provider_id, parent)
      end
      
      protected
      
    end
  end
end