require 'socket'

module Weaver
  module Deployers
    module Execution   
      # This class takes care of setting up the nodes of a root model so that they can run continuous validation
      # rules after deployment.
      #
      # Author:: Fabio Oliveira
      class ContinuousValidationSetup
        include Weaver::Core
        # +root+: Weaver root model
        # +weaver_project_dir+: Directory of the current Weaver project
        # +service_port+: Port on which the Weaver service listens
        def initialize(root, weaver_project_dir, service_port = 9999)
          @root = root
          @weaver_project_dir = weaver_project_dir
          @weaver_service_port = service_port
        end
        
        def add_rules_from_files()
          @@logger.debug "Adding rules from external files to the root model"
          wp = Weaver::Dsl::WeaverParser.new()
          rules_dir = File.join(@weaver_project_dir, 'rules') 
          rules_files = Dir.glob(rules_dir + '/*.weaver')
          @@logger.debug "No rules files found during deployment. Pattern: '#{rules_dir}/*.weaver'" if rules_files.size == 0
          rules_files.each do |file|
            @@logger.debug "Processing #{file}"
             root_statement = Weaver::Dsl::RootStatement.new(file, wp)
             root_statement.model = @root
             wp.parse(file, root_statement)
          end          
        end
        
        # Given the root model used to instantiate the class, this method adds the 'contval' automation to each node that is a target of 
        # one or more continuous validation rules. At deployment time, the 'contval' Chef  automation 
        # will copy the appropriate artifacts to the nodes and configure the appropriate rules to run on them.
        # Note that this method modifies the model.
        def process_continuous_validation_rules_and_artifacts(deployment_id)
          @node_to_rules = compute_rules_targets()
          @node_to_rules.each do |id, rules|
            unless rules.size == 0
              node = @root.node_by_id(id)
              automation = add_contval_automation(node, deployment_id)
              artifact_references = Set.new
              rules.each do |rule|
                @@logger.debug "Adding rule #{rule.id} to contval automation for deployment_id #{deployment_id}"
                add_rule_to_contval_automation(automation, rule)
                artifact_references.merge(rule.contval.artifacts) unless rule.contval.artifacts.nil?
              end

              artifact_references.each do |artifact_id|
                add_artifact_to_contval_automation(automation, @root.artifacts[artifact_id])
              end

              @@logger.debug ":contval Chef automation for node :#{node.id} -> #{automation.inspect}"
              @@logger.debug "  contval.rules = #{automation.amd[:weaver_continuous_validation_rules].value.inspect}"
              @@logger.debug "  contval.artifacts = #{automation.amd[:weaver_continuous_validation_artifacts].value.inspect}"
              
              prepare_automation_for_parameter_propagation(automation)
            end
          end
        end
                
        #   This method inspects the given chef\_node\_attr and transforms it to make sure that each entry _i_ (where _i_ is an integer) 
        # of the form _chef\_node\_attr['weaver']['continuous_validation']['rules'][i]['command'] will contain valid String concatenations 
        # that can be evaluated by Kernel.eval. These transformations need to take place because 'command' may contain 
        # references to Zookeeper in the form of calls to the class Weaver::Coordination::EnvironmentCoordinator. When such references
        # are inserted by Weaver in the middle of a rule command written by the user, the resulting rule command may end up with
        # a value that Kernel.eval cannot parse. 
        def self.format_chef_node_attr(chef_node_attr)
          cont_val_rules = chef_node_attr['weaver']['continuous_validation']['rules'] 
          unless cont_val_rules.nil?
            @@logger.debug "format_chef_node_attr is processing chef_node_attr['weaver']['continuous_validation']['rules']"
            @@logger.debug "  Received: #{cont_val_rules}"
            @@logger.debug "    class: #{cont_val_rules.class}"
            cont_val_rules_hash = eval(cont_val_rules)
            
            cont_val_rules_hash.each do |r|
              rule_command = r[:command]
              @@logger.debug  "Trying to match rule command '#{rule_command}'"
              if rule_command =~ /Weaver::Coordination::Coordinated/
                @@logger.debug "Processing rule command '#{rule_command}'"
                
                rule_command.gsub!(/(\)\.read_data\(.*?\))/) {|match| match.gsub(/\s+/,"") }

                tokens = rule_command.split(' ')
                tokenized_command_str = '"' + tokens[0] + '"'
                for i in 1..tokens.length-1
                  tokenized_command_str += ' + " " + "' + tokens[i] + '"'
                end
                @@logger.debug "Tokenized rule command: '#{tokenized_command_str}'"
                
                tokenized_command_str.gsub!(/"\[Weaver::/, '[Weaver::')
                tokenized_command_str.gsub!(/(\.read_data\(.*?\)\])"/, '\1')
                tokenized_command_str.gsub!(/"Weaver::/, 'Weaver::')
                tokenized_command_str.gsub!(/(\.read_data\(.*?\))"/, '\1')
                tokenized_command_str.gsub!(/^"/, "")
                 
                r[:command] = tokenized_command_str
                @@logger.debug "Final rule command: '#{r[:command]}'"
              end
            end
            
            #new_value = cont_val_rules_hash.to_s.gsub(/(\[Weaver::Coordination::EnvironmentCoordinator.new\(.*?\).join\(.*?\).read\(.*?\)\])/, '\1.join(" ")')
            new_value = cont_val_rules_hash.to_s.gsub(/(\[Weaver::Coordination::CoordinatedTransactionManager.instance.join_transaction\(.*?\).read_data\(.*?\)\])/, '\1.join(" ")')
            new_value.gsub!(/(:command=>.+?read_data\(.+?\))"}/, '\1}')
            new_value.gsub!(/(:command=>.+?read_data\(.+?)\\""}/, '\1"}')
         
            chef_node_attr['weaver']['continuous_validation']['rules'] = new_value
            @@logger.debug "  Produced: #{chef_node_attr['weaver']['continuous_validation']['rules']}"
          end
        end
        
        protected
        
        # Returns a \Hash that associates nodes with the continuous validation rules that apply to them.
        # The \Hash keys are node ids; the values are Arrays of Rules       
        def compute_rules_targets()
#          wp = Weaver::Dsl::WeaverParser.new()
#          rules_dir = File.join(Dir.pwd, 'rules') 
#          rules_files = Dir.glob(rules_dir + '/*.weaver')
#          @@logger.debug "No rules files found during deployment. Pattern: '#{rules_dir}/*.weaver'" if rules_files.size == 0
#          rules_files.each do |file|
#            @@logger.debug "Processing #{file}"
#             root_statement = Weaver::Dsl::RootStatement.new(file, wp)
#             root_statement.model = @root
#             wp.parse(file, root_statement)
#          end

          # Initialize node_to_rules. Associate each potential rule target with an empty Array
          node_to_rules = {}
          @root.flatten_children.select {|c| c.weaver_instance_of?(:node) && c.realized_by.empty? && !Weaver::Transforms::Unroll.template?(c)}.each do |n|
            node_to_rules[n.id] = []
          end                    
          
          # Compute the target node of each rule and update node_to_rules accordingly
          @root.rules.each do |id, r|
            @@logger.debug "Checking rule :#{id} before deployment"
            if r.phase == :continuous
              @@logger.debug "Computing targets of continuous validation rule :#{id}"
              out_messages = []
              targets = r.compute_targets(@root, out_messages)
              unless targets.nil?
                targets.each do |t|
                  if (t.weaver_instance_of?(:node) && t.realized_by.empty? && !Weaver::Transforms::Unroll.template?(t)) 
                    @@logger.info "Associating node :#{t.id} with rule :#{r.id}"
                    node_to_rules[t.id] << r
                  else
                    @@logger.warn "Ignored continuous validation rule target: #{t.inspect}"
                  end
                end
              end
            end
          end
   
          return node_to_rules
        end
        
        # Adds to a node a Chef automation element corresponding to the 'continuous_validation' recipe
        # This Chef automation is structured as follows:
        # chef_automation(:weaver_continuous_validation) { 
        #   role "weaver_continuous_validation"
        #   property :weaver_continuous_validation_artifacts, :name => "weaver.continuous_validation.artifacts"
        #   property :weaver_continuous_validation_rules, :name => "weaver.continuous_validation.rules"
        #   property :weaver_continuous_validation_service_hostname, :name => "weaver.continuous_validation.service_hostname"
        #   property :weaver_continuous_validation_service_port, :name => "weaver.continuous_validation.service_port"
        #   property :weaver_continuous_validation_deployment_id, :name => "weaver.continuous_validation.deployment_id"
        #   property :weaver_continuous_validation_node_id, :name => "weaver.continuous_validation.node_id"
        #} 
        def add_contval_automation(node, deployment_id)
          automation = Weaver::Core::Automation.new(:weaver_continuous_validation, node)
          automation.type = :chef_automation
  
          role = Weaver::Core::Attribute.new(:role, automation)
          role.value = "weaver_continuous_validation"
          automation.amd[:role] = role
            
          deps_file_url = Weaver::Core::Attribute.new(:deps_file_url, automation)
          deps_file_url.name = "DEPS_FILE_URL"
          automation.amd[:deps_file_url] = deps_file_url
          
          attribute = Weaver::Core::Attribute.new(:weaver_continuous_validation_artifacts, automation)
          attribute.name = ("weaver.continuous_validation.artifacts")
          attribute.value = []
          attribute.mutable = true
          automation.amd[attribute.id] = attribute
          
          attribute = Weaver::Core::Attribute.new(:weaver_continuous_validation_rules, automation)
          attribute.name = ("weaver.continuous_validation.rules")
          attribute.value = []
          attribute.mutable = true
          automation.amd[attribute.id] = attribute
          
          attribute = Weaver::Core::Attribute.new(:weaver_continuous_validation_service_hostname, automation)
          attribute.name = ("weaver.continuous_validation.service_hostname")
          attribute.value = Socket.ip_address_list.detect{|i| i.ipv4? && !i.ipv4_loopback? && !i.ipv4_multicast? && !i.ipv4_private?}.ip_address
          attribute.mutable = true
          automation.amd[attribute.id] = attribute
          
          attribute = Weaver::Core::Attribute.new(:weaver_continuous_validation_service_port, automation)
          attribute.name = ("weaver.continuous_validation.service_port")
          attribute.value = @weaver_service_port
          attribute.mutable = true
          automation.amd[attribute.id] = attribute
          
          attribute = Weaver::Core::Attribute.new(:weaver_continuous_validation_deployment_id, automation)
          attribute.name = ("weaver.continuous_validation.deployment_id")
          attribute.value = deployment_id
          attribute.mutable = true
          automation.amd[attribute.id] = attribute
          
          attribute = Weaver::Core::Attribute.new(:weaver_continuous_validation_node_id, automation)
          attribute.name = ("weaver.continuous_validation.node_id")
          attribute.value = node.id
          attribute.mutable = true
          automation.amd[attribute.id] = attribute
          
          node.automation[automation.id] = automation
          
          return automation               
        end
        
        def add_rule_to_contval_automation(contval_automation, rule)
          rule_hash = {}
          rule_hash[:id] = rule.id
          rule_hash[:type] = rule.contval.runtime_type.to_s
          rule_hash[:period] = rule.contval.period
          rule_hash[:command] = rule.contval.command
          
          contval_automation.amd[:weaver_continuous_validation_rules].value << rule_hash
        end
        
        def add_artifact_to_contval_automation(contval_automation, artifact)
          artifact_hash = {}
          artifact_hash[:id] = artifact.id
          artifact_hash[:embedded] = (artifact.location == :local ? true : false)
          artifact_hash[:target] = artifact.target_path
          artifact_hash[:exec] = artifact.exec
          if artifact.location == :local
            artifact_local_path = File.join(@weaver_project_dir, 'rules', 'artifacts', artifact.source_path)
            @@logger.debug "Artifact :#{artifact.id} local path: #{artifact_local_path}"
            file = File.open(artifact_local_path, "r")
            contents = file.read()
            file.close()
            artifact_hash[:artifact] = Base64.strict_encode64(contents)
          end
          #TODO: Deal with non-embedded artifacts
          
          contval_automation.amd[:weaver_continuous_validation_artifacts].value << artifact_hash
        end
        
        def prepare_automation_for_parameter_propagation(automation)
          block_str = automation.amd[:weaver_continuous_validation_rules].value.to_s.gsub("\\#", "#")
          @@logger.debug "Preparing :weaver_continuous_validation automation for parameter propagation"
          @@logger.debug "  rules.to_s = #{automation.amd[:weaver_continuous_validation_rules].value.to_s}"
          
          block = Kernel.eval "Proc.new { #{block_str}.to_s }" 
          @@logger.debug "  Context of LateBinding: #{@root.topologies.inspect}"
          automation.amd[:weaver_continuous_validation_rules].value = LateBinding.new(nil, @root.topologies.values.to_a[0], &block) 
        end
      end
    end
  end
end