require "weaver/core/logging"
require "weaver/core/network_interface_controller"
require "weaver/iwd/iwd_request_generator"
require "weaver/iwd/pattern_utils"
require "weaver/deployers/deployer"
require 'weaver/utils/value_decrypter'
require "devops_deployment_lib/util"
require "devops_deployment_lib/providers/iwd_virtual_system_client" # Ruby client lib to interact with IWD
require "devops_deployment_lib/providers/virtual_system_deployment_manager"
require "devops_deployment_lib/environment"
require 'weaver/providers/iwd_vsp'

module Weaver
  module Deployers
    # A namespace for plugin deployers for \Weaver.
    #
    # @see Plugins::IwdDeployer
    # @see Plugins::MockDeployer
    module Plugins
      # The \IwdDeployer class deploys IWD Virtual \System Pattern instances from a \Weaver
      # infrastructure definition.
      #
      # @see http://www-01.ibm.com/software/webservers/workload-deployer/ IBM Workload Deployer
      class IwdDeployer < Deployer

        #
        # Return and ID for a cloud type that this deployer supports
        #
        def self.id
          return :iwd_vsp
        end 
        
        # IWD specific preprocessing code for the deployer. In this case we need to 
        # add the hostname and IP address macros in so that a redeployment of that
        # topology can correctly propagate the IP and hostname.
        #             
        # *Args*    :
        # * +topology+ the environment topology that needs to  
        def self.serialize_topology_preprocessor(topology) 
          infrastructures = topology.get_cloud_infrastructures()
          raise WeaverError.new("No clouds in environment") if infrastructures.empty?
          infrastructure = infrastructures[0]
          
          infrastructure.unique_nodes.each do |n_id, n|
            Weaver::Iwd::PatternUtils::add_iwd_macros_to_node(infrastructure.provisioner.nil? ? infrastructure.cloud.version : infrastructure.provisioner.version, n)
          end
          topology
        end

        # Creates a new \IwdDeployer for a given target IWD environment.
        # The list of properties for configuration are:
        # * +log_level+ optional log level or Logger::WARN as default
        # * +timeout+ optional otherwise 60*30 sec are used.
        # In addition it expects a key 'iwd_vsp_properties' with at least the following properties
        # * +url+ URL the IWD URL otherwise the one from the Weaver infrastructure definition is used
        # * +username+ the IWD username
        # * +password+ the IWD password
        # * +ssl_verify_cert+ set it to true if SLL certificate of the server should be verified.
        #
        # *Args*    :
        # * +properties+ A list of add properties configuring the deployer.
        def initialize(properties = {}, &reporter)
          super
          @iwd_properties = properties[:iwd_properties] || properties['iwd_properties']
          unless @iwd_properties
            msg = "No 'iwd_properties' element found in YAML properties file (--properties flag in the CLI). Please ensure that
            the following information is present:
               iwd_properties:
                 hostname: YOUR_IWD_URL (e.g., fit-iwd-1.rtp.raleigh.ibm.com)
                 user: YOUR_USER_ID
                 password: YOUR_PWD
                 timeout: 3600 # default
                 ssl_verify_cert: true|false # toogle SSL certificate validation"
            if properties[:simulate]
              @iwd_properties = { :hostname => 'dummy.iwd.ibm.com', :user => 'mock_username', :password => '', :ssl_verify_cert => true }
            else
              raise WeaverError, msg
            end
          end           
          @reporter = reporter || Proc.new { |lvl, status, msg| }
        end

        # Deploy a new instance of a Virtual \System Pattern (VSP) via the DevOps
        # VirtualSystemDeploymentManager (from devops_deployment_lib).
        #
        # *Args*    :
        # * +environment+ An instance of an Topology object that represents the binding
        #   of an application to an environment.
        # * +deploy_parameters+ External attributes that need to be set.
        # *Returns* :
        # * +Environment+ The updated environment instance with runtime information such as hostname and ip.
        #
        def deploy(environment, deploy_parameters = {})
          infrastructures = environment.get_cloud_infrastructures()
          raise WeaverError.new("Multiple clouds in environment: #{infrastructures.map {|inf| inf.id }.inspect}") if infrastructures.size > 1
          raise WeaverError.new("No clouds in environment") if infrastructures.empty?
          infrastructure = infrastructures[0]
          init_deployer(infrastructure.provisioner, infrastructure.cloud)

          infrastructure.amd[:state] = Weaver::Core::Attribute.new(:state, infrastructure) if infrastructure.amd[:state].nil?
            
          begin
            iwd_version = @deployment_manager.vsp_client.describe_version["version"]
            @@logger.debug("IWD version reported = #{iwd_version}")
          rescue StandardError => se
            @@logger.error "Caught #{se.inspect} getting IWD version; backtrace=#{se.backtrace}"
            raise WeaverError.new("Cannot contact virtual system deployment manager", se);
          end
          
          iwd_version.sub!(/-.*$/, "")
          @@logger.debug("IWD version used to determine the right syntax of macros: #{iwd_version}")
          
          generator = Weaver::Iwd::IwdRequestGenerator.new(infrastructure, iwd_version)
          json_request = generator.to_json()
          @@logger.debug("Generating JSON request...")
          # Don't log the JSON request, as it contains passwords
#          @@logger.debug("JSON request: \n#{json_request}")

          # write JSON request to a temp file
          file = Tempfile.new('iwd_request.json')
          begin
            file.write(json_request)
          ensure
            file.close
          end
          @@logger.debug("Sending IWD pattern creation request...")
          raise WeaverError.new("Aborted") if @reporter.call(Logger::INFO, "Deploying", "Sending IWD pattern creation request...") == :abort
          
          ################################################################
          # 1. pre-deployment phase
          
          # TODO: add error handling, timeout, etc.
          #begin
          invoke_services(infrastructure) { |service| service.predeploy(deploy_parameters) }
          Timeout::timeout(300) {
            invoke_services(infrastructure) { |service| service.join }
          }
          #rescue
          #  # terminate weaver 
          #end
          
          begin
            instance = @deployment_manager.deploy(file.path, DevopsDeploymentLib::Environment.new({}))
            @@logger.debug("Received IWD pattern creation response")
          rescue Exceptions::ProvisioningError => pe
            @@logger.error "Caught #{pe.inspect} deploying; backtrace=#{pe.backtrace}"
            record_instance(infrastructure, pe.details)
            raise WeaverError.new("Virtual system deployment failed", pe);
          rescue StandardError => se
            @@logger.error "Caught #{se.inspect} deploying; backtrace=#{se.backtrace}"
            raise WeaverError.new("Virtual system deployment failed", se);
          end
          
          record_instance(infrastructure, instance)
      
          ################################################################
          # 2. deployment phase

          invoke_services(infrastructure) { |service| service.deploy(deploy_parameters) }
        
          raise WeaverError.new("Aborted") if @reporter.call(Logger::INFO, "Deploying", "Received IWD pattern creation response...") == :abort
          @@logger.debug instance

          begin
            instance["virtual_machines"] = @deployment_manager.vsp_client.describe_instance_virtual_machines(instance["id"])
          rescue StandardError => se
            @@logger.error "Caught #{se.inspect} calling describing instance; backtrace=#{se.backtrace}"
            raise WeaverError.new("Virtual system describe instance failed", se);
          end
              
          instance_json = instance["virtual_machines"]
          @@logger.debug instance_json

          @@logger.debug("Fetching additional IWD virtual machine info")
          raise WeaverError.new("Aborted") if @reporter.call(Logger::INFO, "Deploying", "Fetching additional IWD virtual machine info") == :abort
          
          begin
            instance_json_with_key = @deployment_manager.vsp_client.describe_instance_virtual_machines_no_header(instance["id"])
            @@logger.debug instance_json_with_key
          rescue StandardError => se
            @@logger.error "Caught #{se.inspect} calling describing instance phase 2; backtrace=#{se.backtrace}"
            raise WeaverError.new("Virtual system describe instance phase 2 failed", se);
          end
            
          @@logger.debug("Setting instance specific values for environment")
          raise WeaverError.new("Aborted") if @reporter.call(Logger::INFO, "Deploying", "Setting instance specific values for environment") == :abort
          # find the id of the VMs and associate it with node in the infrastructure definition
          id_node_map = {}
          infrastructure.nodes.each do |node_id, node|
            if node_id == node.id # don't include aliases
              node.cloud_status_internal = {}
                
              # Some versions of IWD may send back an 'identifier' in each instance['virtual_machines'] but we 
              # don't count on that.  Instead we will match on the IDs returned from describe_instance_virtual_machines_no_header()
              # with the IDs returned from describe_instance_virtual_machines()
              vm_instance_id = nil
              
              instance_json_with_key.each do |instance|
                if node.name() == instance['partkey'] || node.name() + '_' == instance['partkey']
                  vm_instance_id = instance['id'] 
                  id_node_map[vm_instance_id] = node
                  node.cloud_status_internal.merge!(instance)
                end
              end
              
              unless instance_json.nil? || vm_instance_id.nil?
                instance_json.each do |vmhash|
                  if vm_instance_id == vmhash['id']
                    node.cloud_status_internal.merge!(vmhash)
                  end 
                end
              end

            end
          end

          # query VM ip and hostnames and find the infrastructure node via the id_node_map and set the values
          instance_json.each do |instance|
            node = id_node_map[instance['id']]
              
            if node.nil?
              @@logger.warn "Cannot record #{instance['nics'][0]['ip_hostname']}/#{instance['nics'][0]['ip_address']} for ID #{instance['id']}, expected one of #{id_node_map.keys.inspect}"
            else
              # set the infrastructure node properties
              instance['nics'].each_with_index do |nicinfo, index|
                new_nic = NetworkInterfaceController.new("nic#{index}".to_sym, node)
                new_nic.hostnames << nicinfo['ip_hostname']
                new_nic.ip_address =  nicinfo['ip_address']

                node.nics[new_nic.id] = new_nic
                
                node.amd[:state] = Weaver::Core::Attribute.new(:state, node) if node.amd[:state].nil?
                node.amd[:state].value = node.cloud_status_internal['currentstatus_text']
              end
            end
          end
          
          Timeout::timeout(300) {
            invoke_services(infrastructure) { |service| service.join }
          }
          
          ################################################################
          # 3. post-deployment phase

          invoke_services(infrastructure) { |service| service.postdeploy(deploy_parameters) }
          Timeout::timeout(300) {
            invoke_services(infrastructure) { |service| service.join }
          }
            
          # delete the temp file from above
          file.delete

          return environment
        end
        
        def invoke_services(infrastructure, &block)
          infrastructure.services.each_value { |service| block.call(service) }
          infrastructure.nodes.each do |id, node|
            if node.id == id # Don't handle aliases
              node.services.each_value { |service| block.call(service) }
              node.flatten_children.select { |candidate| candidate.weaver_instance_of?(:component) }.each do |component|
                component.services.each { |service| block.call(service) }
              end
            end
          end           
        end

        # Simulate bringing up the model in IWD
        #
        # *Args*    :
        # * +topology+ A Weaver::Core::Topology that provides a model of an intended deployment and can receive ProblemStatus annotations.
        # * +deploy_parameters+ External attributes that need to be set.
        # *Returns* :
        # * +Topology+ The updated +topology+ with problem annotations and simulated hostname and ip.
        #
        def simulate_deploy(topology, deploy_parameters = {})
          infrastructures = topology.get_cloud_infrastructures()
          if infrastructures.size > 1
            topology.add_validation_failure("Multiple clouds in environment: #{infrastructures.map {|inf| inf.id }.inspect}")
            return topology
          elsif infrastructures.empty?
            topology.add_validation_failure("No clouds in environment")
            return topology
          end
          
          infrastructure = infrastructures[0]
          @@logger.debug "Simulating deployment on infrastructure #{infrastructure.fully_qualified_id}"
          init_deployer(infrastructure.provisioner, infrastructure.cloud)
          
          topology.add_validation_failure("Undefined IWD hostname in provider") if infrastructure.provisioner.host.nil?
          
          # Supply some reasonable-looking IPs and hostnames
          i = 1
          infrastructure.unique_nodes.each do |node_id, node|
            node.ip_address = "192.168.1.#{i}"
            node.hostname = "dyn#{i}.iwd.ibm.com"
            i = i + 1
          end
          
#          # set the virtual_system_id property on the infrastructure
#          vs_id_attr = Attribute.new(:vsys_instance_id, nil)
#          vs_id_attr.name = "IWD Virtual System Instance ID"
#          vs_id_attr.required = false
#          vs_id_attr.value = '1'
#          infrastructure.amd[vs_id_attr.id] = vs_id_attr
          
          # Apply the Chef-specific values
          handler = Weaver::Chef::ChefAutomationHandler.new()
          infrastructure.unique_nodes.each do |node_id, node|
            begin
              # convert each chef_node_attr that is a Hash to a String
              handler.interpolate_and_convert_chef_attributes(infrastructure, node)
              # If there are chef_automations, collect the attribute values to be used together in one chef automation
              # otherwise, attrs will hold default values
              attrs = handler.generate_chef_attributes(node)
              @@logger.debug "Generated Chef node #{attrs[:chef_node].inspect}"
              @@logger.debug "Generated Chef deps_file_url #{attrs[:deps_file_url].inspect}"
              @@logger.debug "Generated (obscured) Chef attr #{attrs[:obscured_chef_node_attr].inspect}"
              
              # UNCOMMENT THE NEXT LINE TO SEE WHAT WILL REALLY BE SENT TO IWD.  THIS WILL
              # EXPOSE PASSWORDS IN THE LOG
              # @@logger.debug "Generated Chef attr #{attrs[:chef_node_attr].inspect}"
            rescue WeaverError => we
              @@logger.warn "Couldn't generate Chef attributes: #{we.message}"
              node.add_validation_failure(we.message)
              attrs = { :chef_node => 'Unknown', :deps_file_url => 'Unknown', :chef_node_attr => ' ' }
            end
            
            node.automation.select { |k, v|
                v.type == :script_package && v.amd[:chef_node] &&
                v.amd[:chef_node_attr] && v.amd[:deps_file_url] }.each do |k, v|
              @@logger.debug "Setting Chef attributes on #{v.fully_qualified_id}"
              # If there is chef automation, it will override chef node values that were specified in the infrastructure 
              # The common case for specifying automation directly in the infrastructure is when a user
              # sets the infrastructure properties in an external yaml file
              if handler.has_chef_automation(node)
                # assign values
                # This may override a value set directly on a chef_node_attr script package automation in the infrastructure
                v.amd[:chef_node].value = attrs[:chef_node]
                v.amd[:chef_node_attr].value = attrs[:chef_node_attr]
                v.amd[:deps_file_url].value = attrs[:deps_file_url]
              else
                # if no chef automation is specified, set the chef script package parameters to default values ONLY if they are nil
                  v.amd[:chef_node].value = v.amd[:chef_node].value || attrs[:chef_node]
                  v.amd[:chef_node_attr].value = v.amd[:chef_node_attr].value ||  attrs[:chef_node_attr]
                  v.amd[:deps_file_url].value = v.amd[:deps_file_url].value || attrs[:deps_file_url]
              end
            end
          end
          
          return topology
        end
       
        private

        # +iwd_provisioner+ Provider of type iwd_vsp
        # +cloud_info+ Cloud (deprecated)
        def init_deployer(iwd_provisioner, cloud_info)
          log_level = @properties[:log_level] || Logger::WARN
          timeout = @properties[:timeout] || 60*30

          url = @iwd_properties[:hostname] || @iwd_properties['hostname'] || iwd_provisioner.url # user URL from properties to allow overriding the one from the infrastructure
          username = @iwd_properties[:user] || @iwd_properties['user']
          password = Weaver::Utils::ValueDecryptor.decrypt_if_needed("PROPS_FILE iwd_properties.password", @iwd_properties[:password] || @iwd_properties['password'])
          ssl_verify_cert = @iwd_properties[:ssl_verify_cert] || @iwd_properties['ssl_verify_cert']

          # create IWD deployment manager
          @deployment_manager = DevopsDeploymentLib::VirtualSystemDeploymentManager.new(
          url, username, password, :log_level => log_level, :timeout => timeout)
          if !ssl_verify_cert
            @deployment_manager.vsp_client.ssl_verify_mode = OpenSSL::SSL::VERIFY_NONE
          end
          @iwd_version = iwd_provisioner.version
        end
        
        def record_instance(infrastructure, instance)
          vsys_instance_id = instance["id"]
            
          # set the virtual_system_id property on the infrastructure
          vs_id_attr = Attribute.new(:vsys_instance_id, nil)
          
#         The name should not be set to prevent validation of the integrated pattern from detecting a new attribute.
#         Unnamed attributes are not checked.
#         vs_id_attr.name = "IWD Virtual System Instance ID"
          vs_id_attr.required = false
          vs_id_attr.value = vsys_instance_id
          infrastructure.amd[vs_id_attr.id] = vs_id_attr
          
          infrastructure.amd[:state].value = instance['currentstatus_text']   
					infrastructure.cloud_status_internal = instance.clone
        end
      end
    end
  end
end 