require 'weaver/iwd/pattern_utils'
require 'weaver/core/status_constants'

module Weaver
  module Iwd
    # Compares a current IWD pattern to an updated IWD pattern that has been updated
    # from the IWD server and collects messages that report the differences.
    # The report indicates if a difference is an error, a warning or informational
    #
    # ===Compares current and updated IWD patterns
    #
    # Author:: Tova Roth
    #
    class IwdPatternUpdateComparator
      include Weaver::Core

      attr_reader :current
      attr_reader :updated
      attr_reader :reporter
      attr_reader :is_predeployment
      # Initialized the current infrastructure, the updated infrastructure and the reporter object.
      #
      # * *Args*    :
      #   - +current+ -> the current infrastructure instance
      #   - +updated+ -> the updated infrastructure instance
      #   - +reported+ -> the base object that will collect the report. If nil the report will be collected on the current pattern
      #
      def initialize(current, updated, reporter = nil, is_predeployment = false)
        @current = current
        @updated = updated
        @reporter = reporter
        @reporter = @current if reporter.nil?
        @is_predeployment = is_predeployment
      end

      # Compares the current pattern against the upated pattern and collects a report on the reporter object
      #
      # * *Args*    :

      def compare_patterns()
        compare_properties()
        compare_image_ids()
        compare_part_ids_to_image_ids()
        compare_parts()
      end

      def compare_properties()

        # NOTE: the validation that compares property lists  on a pattern may not be necessary. Property lists on a pattern are
        # created by Weaver. They are not part of the IWD pattern definition.
        
        pattern_id = @current.amd[:pattern_id].nil? ? "<none>" : @current.amd[:pattern_id].value
        compare_base_properties(@current, @updated, pattern_id, :pattern_property_not_on_server, :pattern_property_added_to_server, :required_pattern_property_added_to_server)

        # TODO compare valid_values of properties for warnings (especially cloud_group, environment_profile, automation_order)

      end


      def compare_part_ids_to_image_ids()

        # get a list of current and updated nodes to compare
        current_nodes_list = PatternUtils.get_unrealized_nodes_list(@current)
        updated_nodes_list = PatternUtils.get_unrealized_nodes_list(@updated)
        
        # loop through each list of nodes, comparing part_id/image_id on both lists. 
        # At the end of the iteration, each list will contain nodes that have a part_id/image_id combination that is not in
        # the other list.
        current_nodes_list.reject! {|current|
          reject_current = false
          updated_nodes_list.reject! { |updated|
            reject_updated = false
            if (!reject_current && updated.part_id == current.part_id && updated.image_id == current.image_id)
              reject_current = true
              reject_updated= true
            end
            reject_updated
          }
          reject_current
        }
       
        # If one list has an image_id that is not in the other list, there is no need to report the error here.
        # That discrepency has already been reported by compare_image_ids. Only report a node if the image_id is
        # in both lists(with different part_ids)
        current_nodes_list.each do |current|
          reject_current = false
          updated_nodes_list.reject! { |updated|
            reject_updated = false
            if (!reject_current && updated.image_id == current.image_id)
              reject_current = true
              reject_updated= true
              msg = Weaver::IWDUtils::MESSAGES[:part_id_to_image_id_changed][:label] % ["#{current.part_name}","#{current.description}","#{current.image_id}", "#{current.part_id}"]
              sev_type = (@is_predeployment ? :predeployment_severity : :severity)
              severity = Weaver::IWDUtils::MESSAGES[:part_id_to_image_id_changed][sev_type]
              report_status(severity, msg)           
            end
            reject_updated
          }
        end

      end
      
      
      def compare_image_ids()

        # make a copy of each hash and look for discrepancies in the list of parts
        current_dup = PatternUtils.get_unrealized_nodes(@current)
        updated_dup =  PatternUtils.get_unrealized_nodes(@updated)

        # iterate through the current list of parts; If both lists have the same image id, delete it from both lists
        # At the end of the iteration, each list will contain image_ids that it has but the other does not
        # This may include an image_id that appears twice in one list and once in the other
        current_dup.each  do | k, v |
          image_id = v.image_id
          updated_key = PatternUtils.get_image_id_key(updated_dup, image_id)
          unless (updated_key.nil?)
            current_dup.delete(k)
            updated_dup.delete(updated_key)
          end
        end

        pattern_id = @current.amd[:pattern_id].nil? ? "<none>" : @current.amd[:pattern_id].value
        current_dup.each do | k, v |
          msg = Weaver::IWDUtils::MESSAGES[:part_not_on_server][:label] % ["#{pattern_id}","#{v.part_name}","#{v.part_id}"]
          sev_type = (@is_predeployment ? :predeployment_severity : :severity)
          severity = Weaver::IWDUtils::MESSAGES[:part_not_on_server][sev_type]
          report_status(severity, msg)
        end

        updated_dup.each do | k, v |
          msg = Weaver::IWDUtils::MESSAGES[:part_added_to_server][:label] % ["#{pattern_id}","#{v.part_name}","#{v.part_id}"]
          sev_type = (@is_predeployment ? :predeployment_severity : :severity)
          severity = Weaver::IWDUtils::MESSAGES[:part_added_to_server][sev_type]
          report_status(severity, msg)
        end

      end


      def compare_parts()
        # since a part may be in the pattern more than once, only process it once
        processed = []
        current_nodes = PatternUtils.get_unrealized_nodes(@current)
        updated_nodes = PatternUtils.get_unrealized_nodes(@updated)
        current_nodes.each do | k, v |
          part_id = v.part_id
          if (!processed.index(part_id))
            processed << part_id
            current_node = PatternUtils.get_part(current_nodes, part_id)
            updated_node = PatternUtils.get_part(updated_nodes, part_id)

            if (current_node && updated_node)
              #compare part properties
              compare_base_properties(current_node, updated_node, part_id, :part_property_not_on_server, :part_property_added_to_server, :required_part_property_added_to_server)
              #             puts ">>>>>>>>>>>>>>>>>>>>>>> Comparing current_node #{current_node.part_id} to updated node #{updated_node.part_id}"
              compare_script_package_ids(current_node, updated_node)
              compare_ids_in_pattern_to_script_ids(current_node, updated_node)
              compare_script_packages(current_node, updated_node)
            end
          end
        end
      end

      # Compares the properties of two base objects and reports discrepancies
      #
      # * *Args*    :
      #   - +current_base+ -> the base object of the current pattern
      #   - +updated_base+ -> the base object of the updated pattern
      #   - +current_base_id+ -> the id of the base object, used in the message
      #   - +deleted_peroperty_msg_id+ -> the message to print when a property has been deleted
      #   - +added_property_msg_id+ -> the message to print when a property has been added
      #   - +added_required_property_msg_id+ -> the message to print when a required property has been added
      # * *Returns* :
      #   - An iwd_node (Node) or nil
      #
      def compare_base_properties(current_base, updated_base, current_base_id, deleted_property_msg_id, added_property_msg_id, added_required_property_msg_id)
         current_dup = current_base.amd.clone || {}
         updated_dup = updated_base.amd.clone || {}
 
         # Delete the intersection of both lists of properties from the lists:
         # Iterate through the current list of properties; If both lists have the same property, delete it from both lists
         # At the end of the iteration, each list will contain properties that it has but the other does not
         # Same properties are determined by both name and group: the name of a property may not be unique, but the name
         # + group combination will be unique. 
         # If the property has no name, it is ommitted from the comparison, for these reasons: All of the properties that are specified
         # in the IWD server have a name.Those are the properties that need to be compared. Other properties are implicit Weaver properties.
         # To avoid problems in which an implicit property later becomes explicit (after folding, for example), don't compare properties
         # without a name.
 
         current_dup.each  do | k, v |
           # if the name is nil, compare by key
           if v.name.nil?
             current_dup.delete(k)
             updated_property = updated_dup[k]
             unless (updated_property.nil?)
               updated_dup.delete(k)
             end
             # if the name is not nil, compare by name + group
           else
             name = v.name
             group = v.group
             updated_key = PatternUtils.get_property_key(updated_dup, name, group)
             unless (updated_key.nil?)
               current_dup.delete(k)
               updated_dup.delete(updated_key)
               #              puts "Yes, deleted from both"
             end
           end
         end
 
         current_dup.each do | k, v |
           msg = Weaver::IWDUtils::MESSAGES[deleted_property_msg_id][:label] %["#{v.name}","#{k}", "#{current_base_id}"]
           sev_type = (@is_predeployment ? :predeployment_severity : :severity)
           severity = Weaver::IWDUtils::MESSAGES[deleted_property_msg_id][sev_type]
           report_status(severity, msg)
         end
 
         updated_dup.each do | k, v |
           unless v.name().nil?
             if (v.mutable && v.required && v.value == nil)
               msg = Weaver::IWDUtils::MESSAGES[added_required_property_msg_id][:label] %["#{v.name}","#{k}", "#{current_base_id}"]
               sev_type = (@is_predeployment ? :predeployment_severity : :severity)
               severity = Weaver::IWDUtils::MESSAGES[added_required_property_msg_id][sev_type]
             else
               msg = Weaver::IWDUtils::MESSAGES[added_property_msg_id][:label] %["#{v.name}","#{k}", "#{current_base_id}"]
               sev_type = (@is_predeployment ? :predeployment_severity : :severity)
               severity = Weaver::IWDUtils::MESSAGES[added_property_msg_id][sev_type]
             end
             report_status(severity, msg)
           end
         end
 
         # TODO compare valid_values of properties and report warnings if changed (especially cloud_group, environment_profile, automation_order)
       end

      def compare_script_package_ids(current_node, updated_node)

        # make a copy of each hash to look for discrepancies in the list of script packages
        current_dup = current_node.automation.dup.reject { |id, a| a.type != :script_package }
        updated_dup = updated_node.automation.dup.reject { |id, a| a.type != :script_package }

        #        puts "BEFORE current_dup.size is  #{current_dup.size}"
        #        puts "BEFORE updated_dup.size is  #{updated_dup.size}"

        # iterate through the current list of script_packages; If both lists have the same script_package id, delete it from both lists
        # At the end of the iteration, each list will contain script_packages that it has but the other does not
        # This may include a script_package that appears twice in one list and once in the other
        current_dup.each  do | k, v |
          script_id = v.script_id
          updated_key = PatternUtils.get_script_package_key(updated_dup, script_id)
          #          puts "Is this script id in both? #{v.name}  #{script_id}"
          unless (updated_key.nil?)
            current_dup.delete(k)
            updated_dup.delete(updated_key)
            #              puts "Yes, deleted from both"
          end
        end

        #        puts "AFTER current_dup.size is  #{current_dup.size}"
        #        puts "AFTER updated_dup.size is  #{updated_dup.size}"

        current_dup.each do | k, v |
          msg = Weaver::IWDUtils::MESSAGES[:script_package_not_in_part][:label] % ["#{v.name}","#{v.script_id}", "#{current_node.part_name}", "#{current_node.part_id}"]
          sev_type = (@is_predeployment ? :predeployment_severity : :severity)
          severity = Weaver::IWDUtils::MESSAGES[:script_package_not_in_part][sev_type]
          report_status(severity, msg)
        end

        updated_dup.each do | k, v |
          msg = Weaver::IWDUtils::MESSAGES[:script_package_added_to_part][:label] % ["#{v.name}","#{v.script_id}", "#{current_node.part_name}", "#{current_node.part_id}"]
          sev_type = (@is_predeployment ? :predeployment_severity : :severity)
          severity = Weaver::IWDUtils::MESSAGES[:script_package_added_to_part][sev_type]
          report_status(severity, msg)
        end

      end
      
      
      def compare_ids_in_pattern_to_script_ids(current_node, updated_node)
        
        # get a list of current and updated script packages to compare
        current_script_pkgs_list = current_node.automation.dup.reject { |id, a| a.type != :script_package }.values
        updated_script_pkgs_list = updated_node.automation.dup.reject { |id, a| a.type != :script_package }.values

        # loop through each list of script packages, comparing id_in_pattern/script_id on both lists.
        # At the end of the iteration, each list will contain script patterns that have a id_in_pattern/script_id combination that is not in
        # the other list.
        current_script_pkgs_list.reject! {|current|
          reject_current = false
          updated_script_pkgs_list.reject! { |updated|
            reject_updated = false
            if (!reject_current && updated.id_in_pattern == current.id_in_pattern && updated.script_id == current.script_id)
              reject_current = true
              reject_updated= true
            end
            reject_updated
          }
          reject_current
        }

        # If one list has an script_id that is not in the other list, there is no need to report the error here.
        # That discrepency has already been reported by compare_script_package_ids. Only report a node if the script_id is
        # in both lists(with different id_in_patterns)
        current_script_pkgs_list.each do |current|
          reject_current = false
          updated_script_pkgs_list.reject! { |updated|
            reject_updated = false
            if (!reject_current && updated.script_id == current.script_id)
              reject_current = true
              reject_updated= true
              msg = Weaver::IWDUtils::MESSAGES[:script_id_to_global_id_changed][:label] % ["#{current.name}","#{current.description}","#{current.script_id}", "#{current.id_in_pattern}"]
              sev_type = (@is_predeployment ? :predeployment_severity : :severity)
              severity = Weaver::IWDUtils::MESSAGES[:script_id_to_global_id_changed][sev_type]
              report_status(severity, msg)
            end
            reject_updated
          }
        end

      end


      def compare_script_packages(current_node, updated_node)
        # since a script package may be in the part more than once, only process it once
        processed = []
        current_node_dup = current_node.automation.dup.reject { |id, a| a.type != :script_package }
        current_node_dup.each do | k, v |
          script_id = v.script_id
          if v.type == :script_package
            if (!processed.index(script_id))
              processed << script_id
              current_script = v
              updated_script = PatternUtils.get_script_package(updated_node.automation, script_id)
              if (current_script && updated_script)
                #compare script parameters
                compare_base_properties(current_script, updated_script, script_id, :script_package_property_not_on_server, :script_package_property_added_to_server, :required_script_package_property_added_to_server)
              end
            end
          end
        end
      end
      
      def report_status(severity, msg)
        label = Weaver::IWDUtils::RESOLUTION_LABELS[:reimport_pattern]
        filename = @current.weaver_origin
        # strip the line number of weaver_origin
        if !filename.nil?
          index = filename.rindex(":")
          if index && index > 0
            filename = " into '" + filename[0..(index-1)] + "'"
          end
        else
          filename = ""
        end

        pattern_id = @current.amd[:pattern_id].nil? ? "<none>" : @current.amd[:pattern_id].value
        description = Weaver::IWDUtils::MESSAGES[:reimport_pattern][:label] % ["#{@current.name}","#{pattern_id}", filename]
        resolutions = []
        resolutions << Resolution.new(label, description, Weaver::Core::ResolutionTypes::REIMPORT_IWD)
        @reporter.report(severity, msg, @current, nil, nil, nil, resolutions)
      end

    end # class
  end # IWD
end # Weaver
