require 'weaver/coordination/logging'
require 'weaver/coordination/system'
require 'weaver/coordination/coordinated_transaction_manager'
require 'weaver/coordination/zk_shared_memory'

module Weaver
  module Coordination
    
    class CoordinatedTransaction
      include Weaver::Coordination::Logging
      
      attr_reader :id
      attr_reader :coordinator
      
      # Establish connection to coordinator service (Apache Zookeeper)
      # *Args*
      #   - +coordinators+:: list of coordination (Zookeeper) servers (IP addresses)
      #   - +coordination_port+:: port used by client to connect to coordination servers
      # *Raises*
      #    Exception if cannot connect to the Apache Zookeeper
      def initialize (transaction_id, coordinator, options = {})
        raise CoordinationError.new() if transaction_id.nil?

        @id = transaction_id
        @coordinator = coordinator
        @connection_options = options
      end # def initialize
      
      def begin
        connect # just in case
      	write_data(Weaver::Coordination::TRANSACTION_STATUS_NM, Weaver::Coordination::TRANSACTION_STATUS::STARTED) 
      end # def begin
      
      def complete
        connect # just in case
      	write_data(Weaver::Coordination::TRANSACTION_STATUS_NM, Weaver::Coordination::TRANSACTION_STATUS::COMPLETED)
      	close
      end # def commit
      
      def abort
        connect # just in case
      	write_data(Weaver::Coordination::TRANSACTION_STATUS_NM, Weaver::Coordination::TRANSACTION_STATUS::ABORTED)
      	close
      end # def abort
      
      def delete
        @zksm.rm_rf(transaction_path)
      end # def delete
      
      def write_data (path, value)
        # @@logger.debug "write_data called for #{path} <- #{value}"
        # @@logger.debug "implementing as #{transaction_path(path)} <- #{value}"
        connect # just in case
        @zksm.write(transaction_path(path), value)
      end # def write_data
      
      # Read a value from the coordinator
      # Params:
      # +node+:: identifier of the node where the parameter is defined
      # +operation+:: identifier of the operation to which the parameter belongs
      # +parameter+:: name of the parameter to read
      # +opts+::
      #   * :timeout time alloted for the read to succeed, default = 0 (ie, unlimited)
      #   * :wait_for_change flag indicating that the current data value is not acceptable, default = false
      # +block+:: optional Proc used to evaluate the data for validity
      def read_data (path, opts = {}, &block)
        connect # just in case
        return @zksm.read(transaction_path(path), opts, &block)
      end # def read
      
      # Indicate end of use of a transaction.  This disconnects any open connection to the coordinator.
      # Does not destroy the transaction record.  A typical usage would be:
      #    transaction = Coordinator#join_transaction(transaction_id)
      #    transaction.read_data(path)
      #    transaction.write_data(path, value)
      #    transaction.close()
      # 
      def close ()
        @zksm.zk.close
      end #

      protected
      
      # Create object w/ read/write semantics to the underlying Apache Zookeeper
      def connect ()
        return unless @zksm.nil?
        @zksm = ZKSharedMemory.new(@coordinator.id, @connection_options.merge({:mock => @coordinator.mock?}))
      end # def connect
      
      def transaction_path(path = nil)
        return @coordinator.transaction_path(@id, path)
      end # def transaction_path
      
      # Converts Hash keys that are strings to symbols.
      # Params:
      # +obj+:: the object to convert
      # protected
      def symbolize_keys (obj)
         if obj.kind_of?(Hash) 
            obj.each_with_object({}) do |(k,v), h| h[k.to_sym] = symbolize_keys(v) end
         elsif obj.kind_of?(Array)
           obj.each_with_object([]) do |v, h| h  << symbolize_keys(v) end
         else
            obj
         end
      end # def symbolize_keys

      # Lists deployed environment instances
      def list 
        # @@logger.debug "/#{Weaver::Coordination::ENVIRONMENT_INSTANCES}/"
        return @zksm.zk.children("/#{Weaver::Coordination::ENVIRONMENT_INSTANCES}")
      end # def list
      
      def get_nodes (id)
        nodes = {}
        return {} if !@zksm.zk.exists?("/#{Weaver::Coordination::ENVIRONMENT_INSTANCES}/#{id}")
          
        @zksm.zk.children("/#{Weaver::Coordination::ENVIRONMENT_INSTANCES}/#{id}").each { |node|
            if @zksm.zk.exists?("/#{Weaver::Coordination::ENVIRONMENT_INSTANCES}/#{id}/#{node}/nics") then
              nic = @zksm.zk.children("/#{Weaver::Coordination::ENVIRONMENT_INSTANCES}/#{id}/#{node}/nics").first
              if @zksm.zk.exists?("/#{Weaver::Coordination::ENVIRONMENT_INSTANCES}/#{id}/#{node}/nics/#{nic}/ipaddress") then
                ip = @zksm.zk.get("/#{Weaver::Coordination::ENVIRONMENT_INSTANCES}/#{id}/#{node}/nics/#{nic}/ipaddress").first
                nodes.merge!({node => ip})
              end
            end
        } #
        return nodes
      end # def get_nodes
      
      def gc (force = false) 
        list.each do |env| 
          if force or get_nodes(env).empty?
            @@logger.debug "removing #{env}"
            remove(env) 
          end
        end
      end # def gc
  
  end # class EnvironmentAttributeStore
    
  end # module Coordination
end # module Weaver